#!/usr/bin/perl
#
###############################################################################
#                                                                             #
#                                                                             #
#                                                                             #
#                 M   M   AAA    GGGG  IIIII   CCCC  K   K                    #
#                 MM MM  A   A  G        I    C      K  K                     #
#                 M M M  AAAAA  G GG     I    C      KKK                      #
#                 M   M  A   A  G   G    I    C      K  K                     #
#                 M   M  A   A   GGG   IIIII   CCCC  K   K                    #
#                                                                             #
#                 SSSSS  TTTTT  U   U  DDDD   IIIII   OOO                     #
#                 SS       T    U   U  D   D    I    O   O                    #
#                  SSS     T    U   U  D   D    I    O   O                    #
#                    SS    T    U   U  D   D    I    O   O                    #
#                 SSSSS    T     UUU   DDDD   IIIII   OOO                     #
#                                                                             #
#                                                                             #
#           Image Conversion, Editing, and Composition on the Web             #
#                                                                             #
#                                                                             #
#                           Software Design                                   #
#                             John Cristy                                     #
#                            November 1997                                    #
#                                                                             #
#                                                                             #
#  Copyright (C) 2000 ImageMagick Studio, a non-profit organization dedicated #
#  to making software imaging solutions freely available.                     #
#                                                                             #
#  Permission is hereby granted, free of charge, to any person obtaining a    #
#  copy of this software and associated documentation files ("ImageMagick"),  #
#  to deal in ImageMagick without restriction, including without limitation   #
#  the rights to use, copy, modify, merge, publish, distribute, sublicense,   #
#  and/or sell copies of ImageMagick, and to permit persons to whom the       #
#  ImageMagick is furnished to do so, subject to the following conditions:    #
#                                                                             #
#  The above copyright notice and this permission notice shall be included in #
#  all copies or substantial portions of ImageMagick.                         #
#                                                                             #
#  The software is provided "as is", without warranty of any kind, express or #
#  implied, including but not limited to the warranties of merchantability,   #
#  fitness for a particular purpose and noninfringement.  In no event shall   #
#  ImageMagick Studio be liable for any claim, damages or other liability,    #
#  whether in an action of contract, tort or otherwise, arising from, out of  #
#  or in connection with ImageMagick or the use or other dealings in          #
#  ImageMagick.                                                               #
#                                                                             #
#  Except as contained in this notice, the name of the ImageMagick Studio     #
#  shall not be used in advertising or otherwise to promote the sale, use or  #
#  other dealings in ImageMagick without prior written authorization from the #
#  ImageMagick Studio.                                                        #
#                                                                             #
###############################################################################
#
#  Magick Studio is a Web interface to PerlMagick that can read and write
#  many of the more popular image formats including JPEG, TIFF, PNM, GIF, and
#  Photo CD.  In addition you can interactively resize, rotate, sharpen, color
#  reduce, or add special effects to your image and save the completed work in
#  the same or differing image format.
#
#
use CGI;
use CGI::Carp qw/fatalsToBrowser/;
use strict;

#
# Global variable declarations.
#
my($action, $ContactInfo, $Debug, $DefaultFont, $DiskLimit, $DocumentDirectory,
   $DocumentRoot, $ExpireCache, $ExpireThreshold, $function, %Functions,
   $header, $IconSize, $length, $LoadAverageThreshold, $MaxFilesize,
   $MaxImageArea, $MaxImageExtent, $MaxWorkFiles, $prefix, $q, %seen, $Timeout,
   $timer, $user_agent);

#
# Change these variables to reflect your environment.
#
$DocumentRoot='/var/www/html';  # server root
$DocumentDirectory='/MagickStudio';  # relative to server root
#$ENV{'ftp_proxy'}='http://webproxy-ae.dupont.com/';
#$ENV{'http_proxy'}='http://webproxy-ae.dupont.com/';
$ENV{DISPLAY}="$ENV{REMOTE_HOST}:0" if $ENV{REMOTE_HOST};
$ENV{LD_LIBRARY_PATH}='/usr/lib:/usr/openwin/lib:/usr/local/lib';
$ENV{MAGICK_FONT_PATH}=$DocumentRoot . $DocumentDirectory . "/fonts";
$ENV{PATH}='/bin:/usr/bin:/usr/openwin/bin:/usr/local/bin';
$ENV{TMPDIR}=$DocumentRoot . $DocumentDirectory . "/tmp";
$ContactInfo='magick-users@imagemagick.org';
$Debug=undef;  # no debugging by default
$DefaultFont="Arial";
$DiskLimit=512;  # 512mb disk limit
$ExpireCache='+1h';  # when to expire browser cache
$ExpireThreshold=4*3600;  # work files time to live
$IconSize='80x80>';
$LoadAverageThreshold=25;  # suspend service if load average exceeds threshold
$MaxFilesize=2048;  # max file size in kilobytes
$MaxImageArea=2048;  # max image area in kilobytes (width*height)
$MaxImageExtent=4096;  # max image extent in kilobytes (width*height*frames)
$MaxWorkFiles=1000;  # max number of work files
$Timeout=60;  # timeout value for uploading an image

#
# Annotate image.
#
sub Annotate
{
  use Image::Magick;

  no strict 'refs';

  my($antialias, $density, $fill, $font, $geometry, $gravity, $image, $path,
    $pointsize, $rotate, $scale, $skew_x, $skew_y, $status, $stroke,
    $strokewidth, $text, $translate, $undercolor);

  #
  # Read image.
  #
  $path=Untaint($q->param('Path'));
  chdir($path) || Error('Your image has expired',$path);
  $image=Image::Magick->new;
  $status=$image->Read("$path/MagickStudio.mpc");
  Error($status) if $#$image < 0;
  #
  # Annotate image.
  #
  $antialias='false';
  $antialias='true' if $q->param('Antialias') eq 'on';
  $density='72';
  $density=$q->param('Density') if $q->param('Density');
  $fill='none';
  $fill=$q->param('Fill') if $q->param('Fill');
  getstore(Untaint($q->param('FontURL')),'MagickStudio.ttf')
    unless $q->param('FontURL') eq 'http://';
  $font="\@MagickStudio.ttf" if -e 'MagickStudio.ttf';
  $font=($image->QueryFont($q->param('Font')))[10] unless -e 'MagickStudio.ttf';
  $geometry='+0+0';
  $geometry=$q->param('Geometry') if $q->param('Geometry');
  $gravity=$q->param('Gravity');
  $pointsize=int($q->param('Pointsize'));
  $rotate=0.0;
  $rotate=$q->param('Rotate') if $q->param('Rotate');
  $scale='0.0, 0.0';
  $scale=$q->param('Scale') if $q->param('Scale');
  $stroke='none';
  $stroke=$q->param('Stroke') if $q->param('Stroke');
  $skew_x=0.0;
  $skew_x=$q->param('SkewX') if $q->param('SkewX');
  $skew_y=0.0;
  $skew_y=$q->param('SkewY') if $q->param('SkewY');
  $strokewidth=1;
  $strokewidth=$q->param('StrokeWidth') if $q->param('StrokeWidth');
  $text=$q->param('Name');
  $text=$q->param('Text') if $q->param('Text');
  $translate='0.0, 0.0';
  $translate=$q->param('Translate') if $q->param('Translate');
  $undercolor='none';
  $undercolor=$q->param('Undercolor') if $q->param('Undercolor');
  $image->Annotate(text=>$text,geometry=>$geometry,font=>$font,fill=>$fill,
    stroke=>$stroke,strokewidth=>$strokewidth,undercolor=>$undercolor,
    pointsize=>$pointsize,density=>$density,gravity=>$gravity,
    translate=>$translate,scale=>$scale,rotate=>$rotate,skewX=>$skew_x,
    skewY=>$skew_y,antialias=>$antialias);
  #
  # Write image.
  #
  CreateWorkDirectory(1);
  Header(GetTitle($image));
  $status=$image->Write(filename=>'MagickStudio.mpc');
  Error($status) if "$status";
  ViewForm($image);
}

#
# Annotate image form.
#
sub AnnotateForm
{
  my(@fonts, @GravityTypes, $image);

  @GravityTypes=
  (
    'Center',
    'NorthWest',
    'North',
    'NorthEast',
    'West',
    'East',
    'SouthWest',
    'South',
    'SouthEast'
  );

  #
  # Display annotate form.
  #
  Header(GetTitle(undef));
  print <<XXX;
To <a href="$DocumentDirectory/Annotate.html">annotate</a> your image with
text, enter your text and location below and press <b>annotate</b>.  There are
additional optional attributes below.  Set them as appropriate.
XXX
  ;
  print $q->startform;
  print $q->hidden(-name=>'CacheID'), "\n";
  print $q->hidden(-name=>'SessionID'), "\n";
  print $q->hidden(-name=>'Path'), "\n";
  print $q->hidden(-name=>'ToolType'), "\n";
  print $q->hidden(-name=>'Name'), "\n";
  print $q->hidden(-name=>'Magick'), "\n";
  print "<dl>\n";
  print "<dl>\n";
  print "<dt>Text:\n";
  print '<dd>', $q->textarea(-name=>'Text',-columns=>50,-rows=>3), "<p>\n";
  print "<dd><table cellpadding=\"2\">\n";
  print "<tr>\n";
  print "<th>Offset</th>\n";
  print "<th>Gravity</th>\n";
  print "</tr>\n";
  print "<tr>\n";
  print '<td>', $q->textfield(-name=>'Geometry',-size=>15,-value=>'+0+0'),
    "</td>\n";
  print '<td>', $q->popup_menu(-name=>'Gravity',-values=>[@GravityTypes]),
    "</td>\n";
  print "</tr>\n";
  print '</table><p>';
  print "</dl>\n";
  print "</dl>\n";
  print 'Press to ', $q->submit(-name=>'Action',-value=>'annotate'),
    ' your image or ', $q->reset(-name=>'reset'), " the form.\n";
  print "<p><hr width=\"50%\" size=\"5\"><p>\n";
  print "<center><h4>Optional Annotation Attributes</h4></center>\n";
  print "<dl>\n";
  print "<dd><table cellpadding=\"2\">\n";
  print "<tr>\n";
  print "<th><a href=\"$DocumentDirectory/Color.html\">Fill Color</a></th>\n";
  print "<th><a href=\"$DocumentDirectory/Color.html\">Stroke Color</a></th>\n";
  print "<th><a href=\"$DocumentDirectory/Color.html\">Undercolor</a></th>\n";
  print "</tr>\n";
  print "<tr>\n";
  print '<td>', $q->textfield(-name=>'Fill',-value=>'white',-size=>15),
    "</td>\n";
  print '<td>', $q->textfield(-name=>'Stroke',-value=>'none',-size=>15),
    "</td>\n";
  print '<td>', $q->textfield(-name=>'Undercolor',-value=>'none',-size=>15),
   "</td>\n";
  print "</tr>\n";
  print '</table><p>';
  print "<dd><table cellpadding=\"2\">\n";
  print "<tr>\n";
  print "<th>PointSize</th>\n";
  print "<th>Density</th>\n";
  print "<th>Stroke Width</th>\n";
  print "</tr>\n";
  print "<tr>\n";
  print "<tr>\n";
  print '<td>', $q->textfield(-name=>'Pointsize',-value=>'24',-size=>15),
    "</td>\n";
  print '<td>', $q->textfield(-name=>'Density',-value=>'72',-size=>15),
    "</td>\n";
  print '<td>', $q->textfield(-name=>'StrokeWidth',-value=>'1',-size=>15),
   "</td>\n";
  print "</tr>\n";
  print '</table><p>';
  print "<dd><table cellpadding=\"2\">\n";
  print "<tr>\n";
  print "<th><a href=\"$DocumentDirectory/fonts/\">Font</a></th>\n";
  print "</tr>\n";
  print "<tr>\n";
  $image=new Image::Magick;
  @fonts=$image->QueryFont();
  print '<td>', $q->scrolling_list(-name=>'Font',-values=>[@fonts],-size=>7),
    "<p>\n";
  print "</tr>\n";
  print "<tr>\n";
  print '<td>', $q->textfield(-name=>'FontURL',-value=>'http://',-size=>30),
    "<p>\n";
  print "</tr>\n";
  print '</table><p>';
  print "<dd><table cellpadding=\"2\">\n";
  print "<tr>\n";
  print "<th>Translate</th>\n";
  print "<th>Scale</th>\n";
  print "<th>Rotate</th>\n";
  print "</tr>\n";
  print "<tr>\n";
  print "<tr>\n";
  print '<td>', $q->textfield(-name=>'Translate',-value=>'0.0, 0.0',-size=>15),
    "</td>\n";
  print '<td>', $q->textfield(-name=>'Scale',-value=>'1.0, 1.0',-size=>15),
    "</td>\n";
  print '<td>', $q->textfield(-name=>'Rotate',-value=>'0.0',-size=>15),
   "</td>\n";
  print "</tr>\n";
  print '</table><p>';
  print "<dd><table cellpadding=\"2\">\n";
  print "<tr>\n";
  print "<th>Skew X</th>\n";
  print "<th>Skew Y</th>\n";
  print "</tr>\n";
  print "<tr>\n";
  print "<tr>\n";
  print '<td>', $q->textfield(-name=>'SkewX',-value=>'0.0',-size=>15),
    "</td>\n";
  print '<td>', $q->textfield(-name=>'SkewY',-value=>'0.0',-size=>15),
    "</td>\n";
  print "</tr>\n";
  print '</table><p>';
  print "<dt> Miscellaneous options:\n";
  print '<dd>', $q->checkbox(-name=>'Antialias',-label=>' antialias text.',
    -checked=>'true'),"\n";
  print "</dl>\n";
  print $q->endform, "\n";
  print <<XXX;
XXX
  ;
  Trailer(1);
}

#
# Suspend service during restricted hours.
#
sub CheckStudioHours
{
  my($hour, $url);

  return if $q->remote_addr() eq '208.253.50.75';
  $hour=(localtime)[2];
  return unless ($hour >= 18) && ($hour < 23);
  $url=$q->self_url . "&Action=" . $action;
  Header("Please Stand By...",-refresh=>"600; URL=$url");
  print <<XXX;
<b>ImageMagick Studio</b> has restricted access from 6PM until 11PM EST.
We have temporarily <i>suspended</i> the processing of your image.
Shortly after 11PM, processing automatically continues and your image is
returned.
XXX
  ;
  Trailer(undef);
}

#
# Suspend service if Studio is too busy.
#
sub CheckStudioStatus
{
  my($load_average, $refresh_rate, $url);

  $load_average=GetLoadAverage();
  return unless $load_average > $LoadAverageThreshold;
  $refresh_rate=int(5*$load_average);
  $url=$q->self_url . "&Action=" . $action;
  Header("Please Stand By...",-refresh=>"$refresh_rate; URL=$url",
    -status=>'502 Service Temporarily Overloaded');
  print <<XXX;
<b>ImageMagick Studio</b> is busy serving other requests.  We have temporarily
<i>suspended</i> the processing of your image.  The current studio status is:
<p>
<center>
<table cellpadding=\"2\">
<tr>
  <th>Load Average</th>
  <th>Threshold</th>
</tr>

<tr>
  <td align=center>$load_average</td>
  <td align=center>$LoadAverageThreshold</td>
</tr>
</table>
</center>
<p>
When the studio load average becomes less than the threshold, processing
automatically continues and your image is returned.  The studio status will
be refreshed in $refresh_rate seconds.
XXX
  ;
  Trailer(undef);
}

#
# Choose a MagickStudio tool.
#
sub ChooseTool
{
  my ($function, $tooltype, %Tools);

  %Tools=
  (
    'Input'=>\&Input,
    'Output'=>\&Output,
    'View'=>\&View,
    'Identify'=>\&Identify,
    'Colormap'=>\&Colormap,
    'Resize'=>\&Resize,
    'Transform'=>\&Transform,
    'Enhance'=>\&Enhance,
    'Effects'=>\&Effects,
    'FX'=>\&Effects,
    'Decorate'=>\&Decorate,
    'Annotate'=>\&Annotate,
    'Draw'=>\&Draw,
    'Composite'=>\&Composite,
    'Comment'=>\&Comment
  );

  $tooltype=$q->param('ToolType');
  View() unless defined($tooltype);
  Error('Unable to mogrify image','path is not defined')
    unless $q->param('Path');
  SaveQueryState($q->param('SessionID'),$tooltype);
  $function=$Tools{$tooltype};
  &$function if defined($function);
  Error('Request failed due to malformed query');
}

#
# Colormap image.
#
sub Colormap
{
  use Image::Magick;

  my($colors, $colorspace, $dither, $global_colormap, $image, $measure_error,
    %options, $path, $status);

  #
  # Read image.
  #
  $path=Untaint($q->param('Path'));
  chdir($path) || Error('Your image has expired',$path);
  $image=Image::Magick->new;
  $status=$image->Read("$path/MagickStudio.mpc");
  Error($status) if $#$image < 0;
  #
  # Quantize image.
  #
  grep($options{$_}++,$q->param('Options'));
  $colors=256;
  $colors=$q->param('Colors') if $q->param('Colors');
  $dither='false';
  $dither='true' if $options{'dither'};
  $global_colormap='false';
  $global_colormap='true' if $options{'global colormap'};
  $colorspace='RGB';
  $colorspace='Gray' if $options{'gray'};
  $measure_error='False';
  $measure_error='True' if $options{'measure error'};
  if ($options{'black & white'})
    {
      $colors='2';
      $colorspace='Gray';
    }
  $colorspace='Transparent' if $options{'preserve transparent pixels'};
  if ($options{'map to clipboard'})
    {
      my($colorcube, $filename);

      #
      # Map to the clipboard image.
      #
      $filename=$DocumentRoot . $DocumentDirectory . '/clipboard/' .
        $q->param('SessionID') . '.mpc';
      $colorcube=Image::Magick->new;
      $status=$colorcube->Read($filename);
      Error($status) if $#$colorcube < 0;
      $image->Map(image=>$colorcube);
    }
  if ($options{'netscape color cube'})
    {
      my($colorcube);

      #
      # Map to the Netscape colorcube.
      #
      $colorcube=Image::Magick->new;
      $status=$colorcube->Read('netscape:');
      Error($status) if $#$colorcube < 0;
      $image->Map(image=>$colorcube);
    }
  $image->Segment(colorspace=>$colorspace) if $options{'segment'};
  $image->Quantize(colors=>$colors,dither=>$dither,colorspace=>$colorspace,
    measure_error=>$measure_error);
  #
  # Write image.
  #
  CreateWorkDirectory(1);
  Header(GetTitle($image));
  $status=$image->Write(filename=>'MagickStudio.mpc');
  Error($status) if "$status";
  ViewForm($image);
}

#
# Colormap form.
#
sub ColormapForm
{
  my(@OptionTypes);

  @OptionTypes=
  [
    'dither',
    'gray',
    'segment',
    'measure error',
    'black & white',
    'global colormap',
    'map to clipboard',
    'netscape color cube',
    'preserve transparent pixels'
  ];

  #
  # Display Colormap form.
  #
  Header(GetTitle(undef));
  print <<XXX;
You have a number of options for creating or changing the image
<a href="$DocumentDirectory/Colormap.html">colormap</a>.  You can reduce the
number of colors in your image, dither, or convert to gray colors.  To create
or modify your image's colormap, check one or more options below.  Next, press
<b>quantize</b> to continue.
XXX
  ;
  print $q->startform;
  print $q->hidden(-name=>'CacheID'), "\n";
  print $q->hidden(-name=>'SessionID'), "\n";
  print $q->hidden(-name=>'Path'), "\n";
  print $q->hidden(-name=>'ToolType'), "\n";
  print $q->hidden(-name=>'Name'), "\n";
  print $q->hidden(-name=>'Magick'), "\n";
  print $q->hidden(-name=>'Action',-value=>'quantize'), "\n";
  print "<dl>\n";
  print "<dl>\n";
  print "<dt>Number of colors:\n";
  print '<dd>', $q->textfield(-name=>'Colors',-size=>15,-value=>'256'), "<p>\n";
  print "<dt>Choose from these options:\n";
  print '<dd>', $q->checkbox_group(-name=>'Options',-values=>@OptionTypes,
    -columns=>2,-default=>'dither'), "<p>\n";
  print "</dl>\n";
  print "</dl>\n";
  print 'Press to ', $q->submit(-name=>'Action',-value=>'quantize'),
    ' your image or ', $q->reset(-name=>'reset'), " the form.\n";
  print $q->endform, "\n";
  print <<XXX;
XXX
  ;
  Trailer(1);
}

#
# Comment.
#
sub Comment
{
  use Digest::SHA1;

  no strict 'subs';

  my($digest, $filename, $path, $remote_host);

  local(*DATA);

  umask(002);
  $path=$DocumentRoot . $DocumentDirectory . "/comments";
  $digest=new Digest::SHA1;
  $digest->add($q->remote_addr(),time(),$$);
  $filename=$path . '/' . $digest->hexdigest . '.txt';
  open(DATA,">$filename") || Error('Unable to save your comments',$filename);
  $remote_host='localhost';
  $remote_host=$q->remote_host() if $q->remote_host();
  print DATA "Host: ", GetHostname($remote_host), "\n";
  print DATA "Address: ", $q->remote_addr(), "\n" if $q->remote_addr();
  print DATA "Ident: ", $q->remote_ident(), "\n" if $q->remote_ident();
  print DATA "User: ", $q->remote_user(), "\n" if $q->remote_user();
  print DATA "Name: ", $q->user_name(), "\n" if  $q->user_name();
  print DATA "Agent: ", $q->user_agent(), "\n" if $q->user_agent();
  print DATA "Time: ", scalar(localtime), "\n";
  print DATA "Comment:\n\n", $q->param('Comment'), "\n";
  close(DATA);
  #
  # Show comments.
  #
  $|=1;
  print $q->header();
  print $q->start_html(-title=>"ImageMagick Studio Comment Form",
    -style=>{-src=>"$DocumentDirectory/css/magick.css"},
    -author=>$ContactInfo,-bgcolor=>'#FFFFFF'), "\n";
  print <<XXX;
<br>
<center>
<a href="http://www.networkeleven.com/direct.php?magick_all">
  <img src="$DocumentDirectory/images/promote.png" alt="NetworkEleven"
    align=right width="186" height="52" border="0"></A>
<img alt="ImageMagick Studio" src="$DocumentDirectory/images/magick.png"
  align=bottom width="114" height="118" border="0">
</center>
<p><hr><p>
XXX
  ;
  print "<dl>\n";
  print "<dl>\n";
  print "<dt>You said:<p>\n";
  print '<dd><pre>', $q->param('Comment'), "</pre><p>\n";
  print "<dt>An administrator will review your comment soon.  Thanks.\n";
  print "</dl>\n";
  print "</dl>\n";
  Trailer(1);
}

#
# Comment form.
#
sub CommentForm
{
  #
  # Display Comment form.
  #
  $|=1;
  print $q->header();
  print $q->start_html(-title=>"ImageMagick Studio Comment Form",
    -style=>{-src=>"$DocumentDirectory/css/magick.css"},
    -author=>$ContactInfo,-bgcolor=>'#FFFFFF'), "\n";
  print <<XXX;
<br>
<center>
<a href="http://www.networkeleven.com/direct.php?magick_all">
  <img src="$DocumentDirectory/images/promote.png" alt="NetworkEleven"
    align=right width="186" height="52" border="0"></a>
<img alt="ImageMagick Studio" src="$DocumentDirectory/images/magick.png"
  align=bottom width="114" height="118" border="0">
</center>
<p><hr><p>
If you have a comment or problem with <b>ImageMagick Studio</b>, enter
any details below and press <b>send</b>.  Be sure to include a valid e-mail
address if you require a response.
XXX
  ;
  $q->delete('Comment');
  print $q->startform;
  print $q->hidden(-name=>'CacheID'), "\n";
  print $q->hidden(-name=>'SessionID'), "\n";
  print $q->hidden(-name=>'Path'), "\n";
  print $q->hidden(-name=>'ToolType'), "\n";
  print $q->hidden(-name=>'Name'), "\n";
  print $q->hidden(-name=>'Magick'), "\n";
  print "<dl>\n";
  print "<dl>\n";
  print '<dd>', $q->textarea(-name=>'Comment',-columns=>50,-rows=>10,
    -wrap=>'horizontal'), "<p>\n";
  print "</dl>\n";
  print "</dl>\n";
  print 'Press to ', $q->submit(-name=>'Action',-value=>'send'),
    ' your comment or ', $q->reset(-name=>'reset'), " the form.\n";
  print $q->endform, "\n";
  print <<XXX;
XXX
  ;
  print "<p><hr><p>\n";
  print $q->end_html;
  exit;
}

#
# Composite image.
#
sub Composite
{
  no strict 'refs';

  use Image::Magick;
  use File::Copy;

  my($color, $compose, $composite, $extent, @extents, $filename,
     $gravity, $geometry, $i, $image, $path, $opacity, $rotate, $slice,
     $status, $tile);

  #
  # Read image.
  #
  $path=Untaint($q->param('Path'));
  chdir($path) || Error('Your image has expired',$path);
  $image=Image::Magick->new;
  $status=$image->Read("$path/MagickStudio.mpc");
  Error($status) if $#$image < 0;
  #
  # Read composite image.
  #
  $filename=Untaint($q->param('CompositeURL'));
  $filename=Untaint($q->param('CompositeFile')) if $q->param('CompositeFile');
  $filename=$q->param('CompositeFile') if $q->param('CompositeFile');
  $filename=$DocumentRoot . $DocumentDirectory . '/clipboard/' .
    $q->param('SessionID') if $filename eq 'clipboard:';
  copy($filename,'MagickStudio.dat') ||
    copy(\*$filename,'MagickStudio.dat') ||
      (getstore($filename,'MagickStudio.dat') eq '200') ||
        Error('Unable to composite image',$filename);
  $composite=Image::Magick->new;
  @extents=$composite->Ping("MagickStudio.dat");
  $extent=0;
  for ($i=0; $i < $#extents; $i+=4) { $extent+=$extents[$i]*$extents[$i+1]; }
  Error('Image extents exceeds maximum allowable') if $extent &&
    ($extent > (1024*$MaxImageExtent));
  $status=$composite->Read('MagickStudio.dat');
  Error($status) if $#$composite < 0;
  unlink('MagickStudio.dat');
  #
  # Composite image.
  #
  if (($#$image == 0) && ($#$composite > 0))
    {
      my($delay);

      #
      # Composite an animation on a background image.
      #
      $i=$#$composite-$#$image+1;
      $image=$image->Morph(frames=>$i);
      for ($i=0; $image->[$i]; $i++)
      {
        $delay=$composite->[$i]->Get('delay');
        $image->Set(delay=>$delay) if $delay;
      }
    }
  $color='none';
  $color=$q->param('BackgroundColor') if $q->param('BackgroundColor');
  $compose=$q->param('ComposeType');
  $geometry='+0+0';
  $geometry=$q->param('Geometry') if $q->param('Geometry');
  $gravity='NorthWest';
  $gravity=$q->param('Gravity') if $q->param('Gravity');
  $rotate=0.0;
  $rotate=$q->param('Rotate') if $q->param('Rotate');
  $tile=0;
  $tile=$q->param('Tile') eq 'on' if $q->param('Tile');
  for ($i=0; $image->[$i]; $i++)
  {
    $slice=$composite->[$i % ($#$composite+1)];
    $image->[$i]->Composite(compose=>$compose,image=>$slice,gravity=>$gravity,
      geometry=>$geometry,rotate=>$rotate,color=>$color,tile=>$tile);
  }
  #
  # Write image.
  #
  CreateWorkDirectory(1);
  Header(GetTitle($image));
  $status=$image->Write(filename=>'MagickStudio.mpc');
  Error($status) if "$status";
  ViewForm($image);
}

#
# Composite image form.
#
sub CompositeForm
{
  my($action, @ComposeTypes, @GravityTypes);

  @ComposeTypes=
  (
    'Over',
    'In',
    'Out',
    'Atop',
    'Xor',
    'Plus',
    'Minus',
    'Add',
    'Subtract',
    'Difference',
    'Multiply',
    'Bumpmap',
    'Copy',
    'CopyRed',
    'CopyGreen',
    'CopyBlue',
    'CopyOpacity',
    'Clear',
    'Dissolve',
    'Displace',
    'Modulate',
    'Threshold'
  ),
  @GravityTypes=
  (
    'Center',
    'NorthWest',
    'North',
    'NorthEast',
    'West',
    'East',
    'SouthWest',
    'South',
    'SouthEast',
  );

  #
  # Composite image form.
  #
  Header(GetTitle(undef));
  print <<XXX;
To <a href="$DocumentDirectory/Composite.html" target="help">composite</a> your
image, press <b>Browse</b> and select your image file or enter the
Uniform Resource Locator of your image.  Next, choose the location of the
composite image and the type of composite operation.  Finally, press
<b>composite</b> to continue.
XXX
  ;
  $action=$q->script_name() . "?CacheID=" . $q->param('CacheID') .
    "&Action=composite";
  print $q->start_multipart_form(-action=>$action);
  print $q->hidden(-name=>'SessionID'), "\n";
  print $q->hidden(-name=>'Path'), "\n";
  print $q->hidden(-name=>'ToolType'), "\n";
  print $q->hidden(-name=>'Name'), "\n";
  print $q->hidden(-name=>'Magick'), "\n";
  print "<dl>\n";
  print "<dl>\n";
  print "<dt><a href=\"$DocumentDirectory/Filename.html\" target=\"help\">",
    "Filename</a>:\n";
  print '<dd>', $q->filefield(-name=>'CompositeFile',-size=>50,
    -maxlength=>1024), "<p>\n";
  print "<dt><a href=\"$DocumentDirectory/URL.html\" target=\"help\">",
    "URL</a>:\n";
  print '<dd>', $q->textfield(-name=>'CompositeURL',-size=>50), "<p>\n";
  print "<dd><table cellpadding=\"2\">\n";
  print "<tr>\n";
  print "<th>Offset</th>\n";
  print "<th>Gravity</th>\n";
  print "</tr>\n";
  print "<tr>\n";
  print '<td>', $q->textfield(-name=>'Geometry',-size=>15,-value=>'+0+0'),
    "</td>\n";
  print '<td>', $q->popup_menu(-name=>'Gravity',-values=>[@GravityTypes]),
    "</td>\n";
  print "</tr>\n";
  print '</table><p>';
  print "</dl>\n";
  print "</dl>\n";
  print 'Press to ', $q->submit(-name=>'Action',-value=>'composite'),
    ' your image or ', $q->reset(-name=>'reset'), " the form.\n";
  print "<p><hr width=\"50%\" size=\"5\"><p>\n";
  print "<center><h4>Optional Composite Attributes</h4></center>\n";
  print "<dl>\n";
  print "<dd><table cellpadding=\"2\">\n";
  print "<tr>\n";
  print "<th>Blend</th>\n";
  print "<th>Compose</th>\n";
  print "</tr>\n";
  print "<tr>\n";
  print '<td>', $q->textfield(-name=>'Blend',-size=>15,-value=>'0%'),
    "</td>\n";
  print '<td>', $q->popup_menu(-name=>'ComposeType',-values=>[@ComposeTypes]),
    "</td>\n";
  print "</tr>\n";
  print '</table><p>';
  print "<dd><table cellpadding=\"2\">\n";
  print "<tr>\n";
  print "<th>Rotate</th>\n";
  print "<th><a href=\"$DocumentDirectory/Color.html\" target=\"help\">",
    "Background Color</a></th>\n";
  print "</tr>\n";
  print "<tr>\n";
  print '<td>', $q->textfield(-name=>'Rotate',-size=>15,-value=>'0.0'),
    "</td>\n";
  print '<td>', $q->textfield(-name=>'BackgroundColor',-value=>'none',
    -size=>15), "</td>\n";
  print "</tr>\n";
  print '</table><p>';
  print "<dt> Miscellaneous options:\n";
  print '<dd> ', $q->checkbox(-name=>'Tile',
    -label=>' tile across and down the image canvas.'),"\n";
  print "</dl>\n";
  print $q->endform, "\n";
  print <<XXX;
XXX
  ;
  Trailer(1);
}
#
# Create a temporary work area for image files.
#
sub CreateWorkDirectory
{
  use Digest::SHA1;

  my($check) = @_;

  my($digest, $path);

#  CheckStudioHours();
  CheckStudioStatus() if $check;
  $path=$DocumentRoot . $DocumentDirectory . '/workarea';
  chdir($path) || Error('Your image has expired',$path);
  umask(002);
  $digest=new Digest::SHA1;
  $digest->add($q->remote_addr(),time(),$$);
  $path.='/' . $digest->hexdigest;
  $ENV{TMPDIR}=$path;
  mkdir($path,0775);
  chdir($path) || Error('Your image has expired',$path);
  $q->param(-name=>'Path',-value=>$path);
  $q->param(-name=>'ToolType',-value=>'View');
}

#
# Decorate image.
#
sub Decorate
{
  use Image::Magick;

  my($color, $geometry, $image, $path, $status);

  #
  # Read image.
  #
  $path=Untaint($q->param('Path'));
  chdir($path) || Error('Your image has expired',$path);
  $image=Image::Magick->new;
  $status=$image->Read("$path/MagickStudio.mpc");
  Error($status) if $#$image < 0;
  #
  # Decorate image.
  #
  $geometry='+0+0';
  $geometry=$q->param('Geometry') if $q->param('Geometry');
  $color='none';
  $color=$q->param('Color') if $q->param('Color');
  $image->Border(geometry=>$geometry,fill=>$color)
    if $q->param('Option') eq 'border *';
  $image->Frame(geometry=>$geometry,fill=>$color)
    if $q->param('Option') eq 'frame *';
  $image->Raise(geometry=>$geometry,raise=>'true')
    if $q->param('Option') eq 'raise *';
  $image->Raise(geometry=>$geometry,raise=>'False')
    if $q->param('Option') eq 'sunken *';
  #
  # Write image.
  #
  CreateWorkDirectory(1);
  Header(GetTitle($image));
  $status=$image->Write(filename=>'MagickStudio.mpc');
  Error($status) if "$status";
  ViewForm($image);
}

#
# Decorate image form.
#
sub DecorateForm
{
  my(@OptionTypes);

  @OptionTypes=
  [
    'border *',
    'frame *',
    'raise *',
    'sunken *'
  ];

  #
  # Display Decorate form.
  #
  Header(GetTitle(undef));
  print <<XXX;
To <a href="$DocumentDirectory/Decorate.html">decorate</a> your image with a
border or frame, set your options below and press <b>decorate</b>.
XXX
  ;
  print $q->startform;
  print $q->hidden(-name=>'CacheID'), "\n";
  print $q->hidden(-name=>'SessionID'), "\n";
  print $q->hidden(-name=>'Path'), "\n";
  print $q->hidden(-name=>'ToolType'), "\n";
  print $q->hidden(-name=>'Name'), "\n";
  print $q->hidden(-name=>'Magick'), "\n";
  print "<dl>\n";
  print "<dl>\n";
  print "<dt>Decoration geometry:\n";
  print '<dd>', $q->textfield(-name=>'Geometry',-size=>20,-value=>'15x15+3+3'),
    "<p>\n";
  print "<dt><a href=\"$DocumentDirectory/Color.html\">Color</a>:\n";
  print '<dd>', $q->textfield(-name=>'Color',-size=>15,-value=>'gray'), "<p>\n";
  print "<dt>Choose from these decorations:\n";
  print '<dd>', $q->radio_group(-name=>'Option',-values=>@OptionTypes,
    -columns=>5,-default=>'frame *'), "<p>\n";
  print "</dl>\n";
  print "</dl>\n";
  print 'Press to ', $q->submit(-name=>'Action',-value=>'decorate'),
    ' your image or ', $q->reset(-name=>'reset'), " the form.\n";
  print $q->endform, "\n";
  print <<XXX;
XXX
  ;
  Trailer(1);
}

#
# Draw image.
#
sub Draw
{
  use Image::Magick;

  my($antialias, $fill, $image, $path, $points, $primitive, $rotate, $scale,
    $skew_x, $skew_y, $status, $stroke, $strokewidth, $translate, $x, $y);

  #
  # Read image.
  #
  $path=Untaint($q->param('Path'));
  chdir($path) || Error('Your image has expired',$path);
  $image=Image::Magick->new;
  $status=$image->Read("$path/MagickStudio.mpc");
  Error($status) if $#$image < 0;
  #
  # Draw image.
  #
  $antialias='false';
  $antialias='true' if $q->param('Antialias') eq 'on';
  $fill='none';
  $fill=$q->param('Fill') if $q->param('Fill');
  $points='+10+10 +60+60';
  $points=$q->param('Coordinates') if $q->param('Coordinates');
  $primitive=$q->param('Primitive');
  $rotate=0.0;
  $rotate=$q->param('Rotate') if $q->param('Rotate');
  $scale='0.0, 0.0';
  $scale=$q->param('Scale') if $q->param('Scale');
  $skew_x=0.0;
  $skew_x=$q->param('SkewX') if $q->param('SkewX');
  $skew_y=0.0;
  $skew_y=$q->param('SkewY') if $q->param('SkewY');
  $stroke='none';
  $stroke=$q->param('Stroke') if $q->param('Stroke');
  $strokewidth=1;
  $strokewidth=$q->param('StrokeWidth') if $q->param('StrokeWidth');
  $translate='0.0, 0.0';
  $translate=$q->param('Translate') if $q->param('Translate');
  $x=0;
  $y=0;
  ($x,$y)=split(/,/,$q->param('Translate')) if $q->param('Translate');
  if (!$q->param('Tile') || ($q->param('Tile') ne 'on'))
    {
      $image->Draw(primitive=>$primitive,fill=>$fill,stroke=>$stroke,
        strokewidth=>$strokewidth,points=>$points,x=>$x,y=>$y,
        translate=>$translate,scale=>$scale,rotate=>$rotate,skewX=>$skew_x,
        skewY=>$skew_y,antialias=>$antialias);
    }
  else
    {
      my($filename, $tile);

      $filename=$DocumentRoot . $DocumentDirectory . '/clipboard/' .
        $q->param('SessionID') . '.mpc';
      $tile=Image::Magick->new;
      $status=$tile->Read($filename);
      Error($status) if $#$tile < 0;
      $image->Draw(primitive=>$primitive,fill=>$fill,stroke=>$stroke,
        strokewidth=>$strokewidth,tile=>$tile,points=>$points,x=>$x,y=>$y,
        translate=>$translate,scale=>$scale,rotate=>$rotate,skewX=>$skew_x,
        skewY=>$skew_y,antialias=>$antialias);
    }
  #
  # Write image.
  #
  CreateWorkDirectory(1);
  Header(GetTitle($image));
  $status=$image->Write(filename=>'MagickStudio.mpc');
  Error($status) if "$status";
  ViewForm($image);
}

#
# Draw image form.
#
sub DrawForm
{
  my(@PrimitiveTypes);

  @PrimitiveTypes=
  (
    'line',
    'rectangle',
    'roundRectangle',
    'arc',
    'ellipse',
    'circle',
    'polyline',
    'polygon',
    'bezier',
    'path',
    'point'
  );

  #
  # Display draw form.
  #
  Header(GetTitle(undef));
  print <<XXX;
To <a href="$DocumentDirectory/Draw.html">draw</a> on your image, choose a
drawing primitive, define it with coordinates, and press <b>draw</b>.  There
are additional optional attributes below.  Set them as appropriate.
XXX
  ;
  print $q->startform;
  print $q->hidden(-name=>'CacheID'), "\n";
  print $q->hidden(-name=>'SessionID'), "\n";
  print $q->hidden(-name=>'Path'), "\n";
  print $q->hidden(-name=>'ToolType'), "\n";
  print $q->hidden(-name=>'Name'), "\n";
  print $q->hidden(-name=>'Magick'), "\n";
  print "<dl>\n";
  print "<dl>\n";
  print "<dt>Primitive:\n";
  print '<dd>', $q->popup_menu(-name=>'Primitive',-values=>[@PrimitiveTypes]),
    "<p>\n";
  print "<dt>Coordinates:\n";
  print '<dd>', $q->textarea(-name=>'Coordinates',-columns=>50,-rows=>2,
    -value=>'+10+10  +60+60',-wrap=>'horizontal'), "<p>\n";
  print "</dl>\n";
  print "</dl>\n";
  print 'Press to ', $q->submit(-name=>'Action',-value=>'draw'),
    ' on your image or ', $q->reset(-name=>'reset'), " the form.\n";
  print "<p><hr width=\"50%\" size=\"5\"><p>\n";
  print "<center><h4>Optional Drawing Attributes</h4></center>\n";
  print "<dl>\n";
  print "<dd><table cellpadding=\"2\">\n";
  print "<tr>\n";
  print "<th><a href=\"$DocumentDirectory/Color.html\">Fill Color</a></th>\n";
  print "<th><a href=\"$DocumentDirectory/Color.html\">Stroke Color</a></th>\n";
  print "<th>Stroke Width</th>\n";
  print "</tr>\n";
  print "<tr>\n";
  print '<td>', $q->textfield(-name=>'Fill',-value=>'white',-size=>15),
    "</td>\n";
  print '<td>', $q->textfield(-name=>'Stroke',-value=>'none',-size=>15),
    "</td>\n";
  print '<td>', $q->textfield(-name=>'StrokeWidth',-size=>15,-value=>'1'),
    "</td>\n";
  print "</tr>\n";
  print '</table><p>';
  print "<dd><table cellpadding=\"2\">\n";
  print "<tr>\n";
  print "<th>Translate</th>\n";
  print "<th>Scale</th>\n";
  print "<th>Rotate</th>\n";
  print "</tr>\n";
  print "<tr>\n";
  print '<td>', $q->textfield(-name=>'Translate',-value=>'0.0, 0.0',-size=>15),
    "</td>\n";
  print '<td>', $q->textfield(-name=>'Scale',-value=>'1.0, 1.0',-size=>15),
    "</td>\n";
  print '<td>', $q->textfield(-name=>'Rotate',-value=>'0.0',-size=>15),
   "</td>\n";
  print "</tr>\n";
  print '</table><p>';
  print "<dd><table cellpadding=\"2\">\n";
  print "<tr>\n";
  print "<th>Skew X</th>\n";
  print "<th>Skew Y</th>\n";
  print "</tr>\n";
  print "<tr>\n";
  print '<td>', $q->textfield(-name=>'SkewX',-value=>'0.0',-size=>15),
    "</td>\n";
  print '<td>', $q->textfield(-name=>'SkewY',-value=>'0.0',-size=>15),
    "</td>\n";
  print "</tr>\n";
  print '</table><p>';
  print "<dt> Miscellaneous options:\n";
  print '<dd> ', $q->checkbox(-name=>'Tile',
    -label=>' paint the drawing primitive with the clipboard image.'),"\n";
  print '<dd>', $q->checkbox(-name=>'Antialias',-label=>' antialias text.',
    -checked=>'true'),"\n";
  print "</dl>\n";
  print $q->endform, "\n";
  print <<XXX;
XXX
  ;
  Trailer(1);
}

#
# Effects image.
#
sub Effects
{
  use Image::Magick;

  my($filename, $image, $parameter, $path, $status);

  #
  # Read image.
  #
  $path=Untaint($q->param('Path'));
  chdir($path) || Error('Your image has expired',$path);
  $image=Image::Magick->new;
  $status=$image->Read("$path/MagickStudio.mpc");
  Error($status) if $#$image < 0;
  #
  # Effects image.
  #
  $parameter=$q->param('Parameter');
  $image->BlackThreshold("$parameter") if
    $q->param('Option') eq 'black threshold *';
  $image->Blur("$parameter") if $q->param('Option') eq 'blur *';
  $image->Charcoal("$parameter") if $q->param('Option') eq 'charcoal drawing *';
  if ($q->param('Option') eq 'convolve *')
    {
      my(@coefficients);

      @coefficients=split(/,/,$parameter);
      $image->Convolve(\@coefficients);
    }
  $image->Despeckle() if $q->param('Option') eq 'despeckle';
  $image->Edge("$parameter") if $q->param('Option') eq 'edge detect *';
  $image->GaussianBlur("$parameter") if
    $q->param('Option') eq 'gaussian blur *';
  $image->Implode("$parameter") if $q->param('Option') eq 'implode *';
  if ($q->param('Option') eq 'morph *')
    {
      my($morph_image);

      if (($#$image+1) == 1)
        {
          #
          # Read clipboard image.
          #
          $filename=$DocumentRoot . $DocumentDirectory . '/clipboard/' .
            $q->param('SessionID') . '.mpc';
          $status=$image->Read($filename);
          Error($status) if $#$image < 0;
        }
      $morph_image=$image->Morph(frames=>$parameter);
      if (ref($morph_image))
        {
          #
          # Replace image sequence with morph sequence.
          #
          undef $image;
          $image=$morph_image;
        }
    }
  if ($q->param('Option') eq 'mosaic')
    {
      my($mosaic_image);

      if (($#$image+1) == 1)
        {
          #
          # Read clipboard image.
          #
          $filename=$DocumentRoot . $DocumentDirectory . '/clipboard/' .
            $q->param('SessionID') . '.mpc';
          $status=$image->Read($filename);
          Error($status) if $#$image < 0;
        }
      $mosaic_image=$image->Mosaic();
      if (ref($mosaic_image))
        {
          #
          # Replace image sequence with mosaic sequence.
          #
          undef $image;
          $image=$mosaic_image;
        }
    }
  $image->OilPaint("$parameter") if $q->param('Option') eq 'oil paint *';
  $image->MedianFilter("$parameter") if
    $q->param('Option') eq 'median filter *';
  $image->ReduceNoise("$parameter") if $q->param('Option') eq 'reduce noise *';
  $image->Emboss(geometry=>$parameter) if $q->param('Option') eq 'emboss *';
  $image->Shade(geometry=>$parameter,gray=>'false')
    if $q->param('Option') eq 'shade *';
  $image->Shade(geometry=>$parameter,gray=>'true')
    if $q->param('Option') eq 'gray shade *';
  $image->Sharpen("$parameter") if $q->param('Option') eq 'sharpen *';
  $image->Solarize("$parameter") if $q->param('Option') eq 'solarize *';
  $image->Spread("$parameter") if $q->param('Option') eq 'spread *';
  if ($q->param('Option') eq 'stegano *')
    {
      my($clipboard);

      #
      # Read clipboard image.
      #
      $filename=$DocumentRoot . $DocumentDirectory . '/clipboard/' .
        $q->param('SessionID') . '.mpc';
      $clipboard=Image::Magick->new(size=>"256x256+$parameter");
      $status=$clipboard->Read($filename);
      Error($status) if $#$clipboard < 0;
      $image->Stegano(image=>$clipboard);
    }
  if ($q->param('Option') eq 'stereo')
    {
      my($stereo);

      $stereo=$image->Clone();
      $stereo->Roll('+4+4');
      $image->Stereo(image=>$stereo);
    }
  $image->Swirl("$parameter") if $q->param('Option') eq 'swirl *';
  $image->AdaptiveThreshold("$parameter") if
    $q->param('Option') eq 'adaptive threshold *';
  $image->Threshold("$parameter") if $q->param('Option') eq 'threshold *';
  $image->UnsharpMask("$parameter") if $q->param('Option') eq 'unsharp mask *';
  $image->Wave("$parameter") if $q->param('Option') eq 'wave *';
  $image->WhiteThreshold("$parameter") if
    $q->param('Option') eq 'white threshold *';
  #
  # Write image.
  #
  CreateWorkDirectory(1);
  Header(GetTitle($image));
  $status=$image->Write(filename=>'MagickStudio.mpc');
  Error($status) if "$status";
  ViewForm($image);
}

#
# Effects form.
#
sub EffectsForm
{
  my(@OptionTypes);

  @OptionTypes=
  [
    'sharpen *',
    'blur *',
    'despeckle',
    'emboss *',
    'threshold *',
    'spread *',
    'shade *',
    'gray shade *',
    'edge detect *',
    'median filter *',
    'reduce noise *',
    'gaussian blur *',
    'adaptive threshold *',
    'unsharp mask *',
    'black threshold *',
    'white threshold *'
  ];

  #
  # Display Effects form.
  #
  Header(GetTitle(undef));
  print <<XXX;
To <a href="$DocumentDirectory/Effects.html">effect</a> your image, enter your
effects parameter and method.  Note, only methods denoted with an asterisk
require a parameter value.  Next, press <b>effect</b> to continue.
XXX
  ;
  print $q->startform;
  print $q->hidden(-name=>'CacheID'), "\n";
  print $q->hidden(-name=>'SessionID'), "\n";
  print $q->hidden(-name=>'Path'), "\n";
  print $q->hidden(-name=>'ToolType'), "\n";
  print $q->hidden(-name=>'Name'), "\n";
  print $q->hidden(-name=>'Magick'), "\n";
  print $q->hidden(-name=>'Action',-value=>'effect'), "\n";
  print "<dl>\n";
  print "<dl>\n";
  print "<dt>Parameter:\n";
  print '<dd>', $q->textfield(-name=>'Parameter',-size=>20,-value=>'0.0x1.0'),
    "<p>\n";
  print "<dt>Choose from these effects</a>:\n";
  print '<dd>', $q->radio_group(-name=>'Option',-values=>@OptionTypes,
    -columns=>3), "<p>\n";
  print "</dl>\n";
  print "</dl>\n";
  print 'Press to ', $q->submit(-name=>'Action',-value=>'effect'),
    ' your image or ', $q->reset(-name=>'reset'), " the form.\n";
  print $q->endform, "\n";
  print <<XXX;
XXX
  ;
  Trailer(1);
}

#
# Enhance image.
#
sub Enhance
{
  use Image::Magick;

  my($image, $parameter, $path, $status);

  #
  # Read image.
  #
  $path=Untaint($q->param('Path'));
  chdir($path) || Error('Your image has expired',$path);
  $image=Image::Magick->new;
  $status=$image->Read("$path/MagickStudio.mpc");
  Error($status) if $#$image < 0;
  #
  # Enhance image.
  #
  $parameter=$q->param('Parameter');
  $image->Contrast(sharpen=>'true') if $q->param('Option') eq 'spiff';
  $image->Contrast(sharpen=>'false') if $q->param('Option') eq 'dull';
  $image->Equalize() if $q->param('Option') eq 'equalize';
  $image->Gamma(gamma=>$parameter) if $q->param('Option') eq 'gamma *';
  $image->Level(level=>$parameter) if $q->param('Option') eq 'level *';
  $image->Normalize() if $q->param('Option') eq 'normalize';
  $image->Modulate(hue=>$parameter) if $q->param('Option') eq 'hue *';
  $image->Modulate(saturation=>$parameter)
    if $q->param('Option') eq 'saturation *';
  $image->Modulate(brightness=>$parameter)
    if $q->param('Option') eq 'brightness *';
  $image->Negate() if $q->param('Option') eq 'negate';
  #
  # Write image.
  #
  CreateWorkDirectory(1);
  Header(GetTitle($image));
  $status=$image->Write(filename=>'MagickStudio.mpc');
  Error($status) if "$status";
  ViewForm($image);
}

#
# Enhance form.
#
sub EnhanceForm
{
  my(@OptionTypes);

  @OptionTypes=
  [
    'spiff',
    'dull',
    'equalize',
    'normalize',
    'negate',
    'hue *',
    'saturation *',
    'brightness *',
    'gamma *',
    'level *'
  ];

  #
  # Display Enhancement form.
  #
  Header(GetTitle(undef));
  print <<XXX;
To <a href="$DocumentDirectory/Enhance.html">enhance</a> your image, enter your
enhancement parameter and method.  Note, only methods denoted with an asterisk
require a parameter value.  Next, press <b>enhance</b> to continue.
XXX
  ;
  print $q->startform;
  print $q->hidden(-name=>'CacheID'), "\n";
  print $q->hidden(-name=>'SessionID'), "\n";
  print $q->hidden(-name=>'Path'), "\n";
  print $q->hidden(-name=>'ToolType'), "\n";
  print $q->hidden(-name=>'Name'), "\n";
  print $q->hidden(-name=>'Magick'), "\n";
  print $q->hidden(-name=>'Action',-value=>'enhance'), "\n";
  print "<dl>\n";
  print "<dl>\n";
  print "<dt>Parameter:\n";
  print '<dd>', $q->textfield(-name=>'Parameter',-size=>20,-value=>'1.6'),
    "<p>\n";
  print "<dt>Choose from these enhancements</a>:\n";
  print '<dd>', $q->radio_group(-name=>'Option',-values=>@OptionTypes,
    -columns=>3,-default=>'gamma *'), "<p>\n";
  print "</dl>\n";
  print "</dl>\n";
  print 'Press to ', $q->submit(-name=>'Action',-value=>'enhance'),
    ' your image or ', $q->reset(-name=>'reset'), " the form.\n";
  print $q->endform, "\n";
  print <<XXX;
XXX
  ;
  Trailer(1);
}

#
# Display an error.
#
sub Error
{
  my($message,$qualifier) = @_;

  Header($message) unless $header;
  $qualifier="" unless $qualifier;
  print <<XXX;
<center>
<img src="$DocumentDirectory/images/stop.png" width="48" height="48" border="0">
</center>
<p> <br>
<dl>
<dt><font face="Arial,Helvetica" size=+3>$message:</font>
<p>
<dd><font face="Arial,Helvetica" size=+2>$qualifier</font>
<p>
<dd><font face="Arial,Helvetica" size=+1>$!</font>
</dl>
<br> <p>
<center>
<img src="$DocumentDirectory/images/stop.png" width="48" height="48" border="0">
</center>
<p>
<br>
<br>
Press <b>Back</b> to correct the error or press a tab above to continue.
XXX
  ;
  Trailer(1);
  die $message;
}

#
# Expire files in the workarea that are older than some threshold.
#
sub ExpireFiles
{
  my($path, $age) = @_;

  use File::Path;

  my(@files, $number_files);

  local(*DIR);

  chdir($path) || Error('Your image has expired',$path);
  opendir(DIR,".");
  @files=readdir(DIR);
  closedir(DIR);
  $number_files=$#files+1;
  for (@files)
  {
    next if /^\./;
    next if /^index.html/;
    next if (time()-(stat($_))[9]) < $age;
    rmtree($_,0,1);
    $number_files--;
  }
  ExpireFiles($path,$age/2) if
    ($number_files > $MaxWorkFiles) && ($age > 300);
}

#
# Fetch all the images from an HTTP directory.
#
sub PursueLink
{
  my($prefix, $type, $quote, $url, $base, $depth) = @_;

  $url=url($url,$base)->abs;
  $prefix . $quote . FetchImages($url,$type,$depth) . $quote;
}

sub FetchImages
{
  my($url, $type, $depth) = @_;

  use URI::URL;
  use LWP::UserAgent;
  use LWP::MediaTypes qw(media_suffix);

  my($base, $content, $content_type, $name, $plain_url, $result, $seen,
    $suffix);

  $url=url($url) unless ref($url);
  if ($depth == 0)
    {
      #
      # Limit to URLs below this one.
      #
      $user_agent=new LWP::UserAgent;
      $user_agent->agent('MagickStudio/1.0 ' . $user_agent->agent);
      $user_agent->env_proxy if $ENV{'http_proxy'};
      $prefix=url($url);
      eval
      {
        $prefix->eparams(undef);
        $prefix->equery(undef);
      };
      $_=$prefix->epath;
      s|[^/]+$||;
      $prefix->epath($_);
      $prefix=$prefix->as_string;
      %seen=();
      $length=0;
      FetchImages($url,$type,$depth+1);
      return(undef);
    }
  $type||='a';
  $type='img' if $type eq 'body';
  $depth||=0;
  return($url->as_string) if $url->scheme eq 'mailto';
  $plain_url=$url->clone;
  $plain_url->frag(undef);
  $seen=$seen{$plain_url->as_string};
  if ($seen)
    {
      my($fragment);

      #
      # We have already seen this document.
      #
      $fragment=$url->frag;
      $seen.="#$fragment" if defined($fragment);
      return($seen);
    }
  return($url) if $depth > 2;  # no recursion
  return $url->as_string
    if ($type ne 'img') and ($url->as_string !~ /^\Q$prefix/o);
  #
  # Fetch image.
  #
  $result=$user_agent->request(HTTP::Request->new(GET=>$url));
  if (!$result->is_success)
    {
      $seen{$plain_url->as_string}="*BAD*";
      return("*BAD*");
    }
  $content=$result->content;
  $content_type=$result->content_type;
  #
  # Construct an image name.
  #
  $url=$result->request->url;
  $url=url($url) unless ref($url);
  $name=$url->path;
  $name=~s|.*/||;
  $name=~s|\..*||;
  $name="index" unless length($name);
  $suffix=media_suffix($content_type);
  $name.=".$suffix" if $suffix;
  $seen{$plain_url->as_string}=$name;
  if ($content_type ne "text/html")
    {
      local(*FILE);

      #
      # Save document to disk.
      #
      return($name) if $name eq 'back.gif';
      return($name) if $name eq 'blank.gif';
      return($name) if $name eq 'image2.gif';
      open(FILE,">$name") || Error('Unable to write image file',$name);
      binmode(FILE);
      print FILE $content;
      close(FILE);
      Error('Image size exceeds maximum allowable',$plain_url->as_string)
        unless $length < (1024*$MaxFilesize);
      $length+=$result->content_length;
      return($name);
    }
  #
  # Follow the links in this HTML document.
  #
  $base=$result->base;
  $content=~s/(<\s*(img|a|body)\b[^>]+\b(?:src|href|background)\s*=\s*)(["']?)([^>\s]+)\3/PursueLink($1,lc($2),$3,$4,$base,$depth+1)/gie;       #";
  return($name);
}

#
# Special Effects form.
#
sub FXForm
{
  my(@OptionTypes);

  @OptionTypes=
  [
    'convolve *',
    'implode *',
    'morph *',
    'mosaic',
    'oil paint *',
    'solarize *',
    'stegano *',
    'stereo',
    'swirl *',
    'wave *',
    'charcoal drawing *'
  ];

  #
  # Display Effects form.
  #
  Header(GetTitle(undef));
  print <<XXX;
To add special <a href="$DocumentDirectory/FX.html">effects</a> to your image,
enter your effects parameter and method.  Note, only methods denoted with an
asterisk require a parameter value.  Next, press <b>effect</b> to continue.
XXX
  ;
  print $q->startform;
  print $q->hidden(-name=>'CacheID'), "\n";
  print $q->hidden(-name=>'SessionID'), "\n";
  print $q->hidden(-name=>'Path'), "\n";
  print $q->hidden(-name=>'ToolType'), "\n";
  print $q->hidden(-name=>'Name'), "\n";
  print $q->hidden(-name=>'Magick'), "\n";
  print $q->hidden(-name=>'Action',-value=>'effect'), "\n";
  print "<dl>\n";
  print "<dl>\n";
  print "<dt>Parameter:\n";
  print '<dd>', $q->textfield(-name=>'Parameter',-size=>20,-value=>'60'),
    "<p>\n";
  print "<dt>Choose from these special effects</a>:\n";
  print '<dd>', $q->radio_group(-name=>'Option',-values=>@OptionTypes,
    -columns=>3,-default=>'swirl *'), "<p>\n";
  print "</dl>\n";
  print "</dl>\n";
  print 'Press to ', $q->submit(-name=>'Action',-value=>'effect'),
    ' your image or ', $q->reset(-name=>'reset'), " the form.\n";
  print $q->endform, "\n";
  print <<XXX;
XXX
  ;
  Trailer(1);
}

#
# Return hostname from host address.
#
sub GetHostname
{
  my($hostname) = @_;

  my($address);

  $address=pack("C4",split(/\./,$hostname));
  $hostname=(gethostbyaddr($address,2))[0];
  $hostname;
}

#
# Return the average number of jobs in the run queue over the last minute.
#
sub GetLoadAverage
{
  my($load_average, $os);

  unless ($os = $^O)
  {
    require Config;
    $os=$Config::Config{'osname'};
  }
  $load_average=0;
  return if $os =~ /Win/i;
  if (!(-e "/proc/loadavg"))
    { $load_average=`uptime` =~ /average:\s+(\S+),/; }
  else
    {
      local(*DATA);

      open(DATA,"/proc/loadavg");
      $load_average=<DATA>;
      $load_average=~s/ .*//;
      close(DATA);
    }
  $load_average;
}

#
# Generate the standard HTML title.
#
sub GetTitle
{
  use Image::Magick;

  my($image) = @_;

  my($height, $title, $width);

  if ($image)
    { ($width,$height)=$image->Get('columns','rows'); }
  else
    {
      my($format, $path, $size);

      $path=Untaint($q->param('Path'));
      $image=Image::Magick->new;
      ($width,$height,$size,$format)=$image->Ping("$path/MagickStudio.mpc");
    }
  $title=$q->param('Name') . '.' . $q->param('Magick') .  '  ' .  $width .
    'x' . $height;
  $title;
}

#
# Print the standard HTML header with the MagickStudio logo.
#
sub Header
{
  my($title, @attributes) = @_;

  my($p, %tools, $tooltype, $url);

  #
  # Initialize tool types.
  #
  $tools{'Input'}='input';
  $tools{'Output'}='output';
  $tools{'View'}='view';
  $tools{'Identify'}='identify';
  $tools{'Colormap'}='colormap';
  $tools{'Resize'}='resize';
  $tools{'Transform'}='transform';
  $tools{'Enhance'}='enhance';
  $tools{'Effects'}='effects';
  $tools{'FX'}='fx';
  $tools{'Decorate'}='decorate';
  $tools{'Annotate'}='annotate';
  $tools{'Draw'}='draw';
  $tools{'Composite'}='composite';
  $q->param(-name=>'ToolType',-value=>'View') unless
    defined($q->param('ToolType'));
  $tooltype=$q->param('ToolType');
  $tools{$tooltype}.='_highlighted';
  #
  # Print the standard HTML header with the MagickStudio logo.
  #
  $header=1;
  $|=1;
  print $q->header(-expires=>$ExpireCache,@attributes), "\n";
  print $q->start_html(-title=>$title,
    -style=>{-src=>"$DocumentDirectory/css/magick.css"},
    -author=>$ContactInfo,-marginheight=>1,-marginwidth=>1,-topmargin=>1,
    -leftmargin=>1), "\n";
  print <<XXX;
<table border="0" cellpadding="0" cellspacing="0" summary="Masthead" width="100%"> <tr>
  <td bgcolor="#003399" width="25%" height="118" border="0" background="$DocumentDirectory/images/background.gif"><a href="http://www.imagemagick.org/"><img src="../images/script.gif" width="278" height="118" border="0" alt=""></a></td>
  <td bgcolor="#003399" width="60%" height="118" border="0" background="$DocumentDirectory/images/background.gif"><a href="http://www.networkeleven.com/direct.php?magick_all"><img src="../images/promote.png" border="0" width="186" height="52" vspace="29" alt="Powered by NetworkEleven"></a></td>
  <td bgcolor="#003399" width="114" height="118" border="0" align="right"><img src="$DocumentDirectory/images/sprite.png" width="114" height="118" border="0" alt=""></td>
  <td bgcolor="#003399" width="114" height="118" border="0" align="right"><a href="http://www.imagemagick.net"><img src="$DocumentDirectory/images/logo.png" width="114" height="118" border="0" alt="ImageMagick logo"></a></td>
</tr></table>
<table align="left" border="0" cellpadding="10" cellspacing="0" width="100%"><tr><td>
XXX
  ;
  $url=$q->script_name;
  $url.='?CacheID=' .  $q->param('CacheID') if $q->param('CacheID');
  $url.='&SessionID=' . $q->param('SessionID') if $q->param('SessionID');
  $url.='&Path=' . $q->param('Path') if  $q->param('Path');
  $url.='&Name=' . $q->param('Name') if  $q->param('Name');
  $url.='&Magick=' . $q->param('Magick') if  $q->param('Magick');
  $url.='&Action=mogrify';
  if (defined($action))
    {
      #
      # Define toolbar.
      #
      print <<XXX;
<table width="100%" border="0" cellspacing="0" cellpadding="0" align="center">
<tr><td align="center">
  <a href="$url&ToolType=Input"> <img width="66" height="21" border="0" vspace="2" src="$DocumentDirectory/images/$tools{'Input'}.png"></a>
  <a href="$url&ToolType=Output"> <img width="76" height="21" border="0" vspace=2 src="$DocumentDirectory/images/$tools{'Output'}.png"></a>
  <a href="$url&ToolType=View"> <img width="67" height="21" border="0" vspace=2 src="$DocumentDirectory/images/$tools{'View'}.png"></a>
  <a href="$url&ToolType=Identify"> <img width="79" height="21" border="0" vspace=2 src="$DocumentDirectory/images/$tools{'Identify'}.png"></a>
  <a href="$url&ToolType=Colormap"> <img width="94" height="21" border="0" vspace=2 src="$DocumentDirectory/images/$tools{'Colormap'}.png"></a>
  <a href="$url&ToolType=Resize"> <img width="74" height="21" border="0" vspace=2 src="$DocumentDirectory/images/$tools{'Resize'}.png"></a>
  <a href="$url&ToolType=Transform"> <img width="93" height="21" border="0" vspace=2 src="$DocumentDirectory/images/$tools{'Transform'}.png"></a>
XXX
      ;
      print "<br>\n" unless $q->user_agent() =~ /WebTV/;
      print <<XXX;
  <a href="$url&ToolType=Enhance"> <img width="85" height="21" border="0" vspace=2 src="$DocumentDirectory/images/$tools{'Enhance'}.png"></a>
  <a href="$url&ToolType=Effects"> <img width="74" height="21" border="0" vspace=2 src="$DocumentDirectory/images/$tools{'Effects'}.png"></a>
  <a href="$url&ToolType=FX"> <img width="59" height="21" border="0" vspace=2 src="$DocumentDirectory/images/$tools{'FX'}.png"></a>
  <a href="$url&ToolType=Decorate"> <img width="88" height="21" border="0" vspace=2 src="$DocumentDirectory/images/$tools{'Decorate'}.png"></a>
  <a href="$url&ToolType=Annotate"> <img width="88" height="21" border="0" vspace=2 src="$DocumentDirectory/images/$tools{'Annotate'}.png"></a>
  <a href="$url&ToolType=Draw"> <img width="69" height="21" border="0" vspace=2 src="$DocumentDirectory/images/$tools{'Draw'}.png"></a>
  <a href="$url&ToolType=Composite"> <img width="97" height="21" border="0" vspace=2 src="$DocumentDirectory/images/$tools{'Composite'}.png"></a>
</td></tr></table>
XXX
      ;
    }
  print <<XXX;
<p><br>
XXX
  ;
}

#
# Identify an image.
#
sub Identify
{
  use Image::Magick;

  my($class, $color, @colors, $filename, $height, $image, $images, $matte,
    $magick, $path, $status, $width, $x);

  #
  # Read image.
  #
  $path=Untaint($q->param('Path'));
  chdir($path) || Error('Your image has expired',$path);
  $images=Image::Magick->new;
  $status=$images->Read("$path/MagickStudio.mpc");
  Error($status) if $#$images < 0;
  #
  # Display image on web page.
  #
  for ($x=0; $images->[$x]; $x++)
  {
    $image=$images->[$x];
    $colors[$x]=$image->Get('colors');
  }
  $magick=$q->param('Magick');
  $filename=$q->param('Name') . '.' . $magick;
  Header(GetTitle($images));
  print <<XXX;
Here is a detailed description of your image, $filename:
<p>
XXX
  ;
  print "<dl>\n";
  print "<dd><pre>\n";
  for ($x=0; $images->[$x]; $x++)
  {
    $image=$images->[$x];
    print "Image number: $x\n";
    $image->Set(filename=>$filename,magick=>$magick);
    $image->Describe();
  }
  print "</pre>\n";
  print "</dl>\n";
  Trailer(1);
}

#
# Input image.
#
sub Input
{
  no strict 'refs';
  no strict 'subs';

  use Image::Magick;
  use LWP::Simple;
  use File::Basename;
  use File::Copy;
  use Digest::SHA1;

  my(@attributes, $basename, $digest, $extent, @extents, $filename, $format,
     $i, $image, $path, $session, $status, $scene);

  #
  # Expire files.
  #
  $path=$DocumentRoot . $DocumentDirectory;
  ExpireFiles("$path/workarea",$ExpireThreshold);
  ExpireFiles("$path/clipboard",$ExpireThreshold);
  ExpireFiles("$path/tmp",$ExpireThreshold);
  ExpireFiles("$path/session_info",7*$ExpireThreshold);
  ExpireFiles("$path/comments",14*$ExpireThreshold);
  #
  # Read image.
  #
  Error('You must specify either an image filename or URL') unless
    $q->param('File') || $q->param('URL') || $q->param('Meta');
  CreateWorkDirectory(undef);
  $path=$q->param('Path');
  $format='';
  $format=$q->param('Format') . ':'
    if ($q->param('Format') && ($q->param('Format') ne 'Implicit'));
  $scene='';
  $scene='[' . $q->param('Scene') . ']' if $q->param('Scene');
  if ($q->param('File'))
    {
      #
      # Copy data file to workarea.
      #
      $filename=Untaint($q->param('File'));
      $filename=$q->param('File');
      copy($filename,'MagickStudio.dat') ||
        copy(\*$filename,'MagickStudio.dat') ||
          getstore($filename,'MagickStudio.dat');
      if (-z 'MagickStudio.dat')
        {
          $q->param('SizeGeometry')=~/(\d+)\D*(\d*)/;
          Error('Image area exceeds maximum allowable') if $1 && $2 &&
            (($1*$2) > (1024*$MaxImageArea));
          Error('Image area exceeds maximum allowable') if $1 && !$2 &&
            (($1*$1) > (1024*$MaxImageArea));
          $image=Image::Magick->new;
          $image->Set(size=>$q->param('SizeGeometry')) if
            $q->param('SizeGeometry');
          $filename=$DocumentRoot . $DocumentDirectory . '/clipboard/' .
            $q->param('SessionID') . '.mpc' if $filename eq 'clipboard:';
          $format=$q->param('Format') . ':' if $q->param('Format');
          $status=$image->Read("$format$filename$scene");
          Error($status) if $#$image < 0;
          $status=$image->Write('MagickStudio.dat');
          $status=$image->Write(filename=>'mpc:MagickStudio.dat') if "$status";
          Error($status) if "$status";
          $format='';
        }
    }
  else
    {
      if ($q->param('URL'))
        {
          $filename=Untaint($q->param('URL'));
          if (!($filename =~ /\/$/))
            {
              #
              # Copy HTTP content to MagickStudio.dat.
              #
              getstore($filename,'MagickStudio.dat');
              if ((-f 'MagickStudio.dat') && (-s 'MagickStudio.dat'))
                {
                  if (($filename =~ /.html$/) || ($filename =~ /.htm$/))
                    {
                      #
                      # Convert HTML content to Postscript.
                      #
                      $image=Image::Magick->new;
                      @extents=$image->Ping("$filename");
                      $extent=0;
                      for ($i=0; $i < $#extents; $i+=4)
                        { $extent+=$extents[$i]*$extents[$i+1]; }
                      Error('Image extents exceeds maximum allowable') if
                        $extent && ($extent > (1024*$MaxImageExtent));
                      $status=$image->Read($filename);
                      Error($status) if $#$image < 0;
                      $status=$image->Write(filename=>'mpc:MagickStudio.dat');
                      Error($status) if "$status";
                    }
                }
            }
          else
            {
              #
              # Copy all images in HTTP directory to MagickStudio.dat.
              #
              FetchImages($filename,undef,0);
              $image=Image::Magick->new;
              @extents=$image->Ping("*");
              $extent=0;
              for ($i=0; $i < $#extents; $i+=4)
                { $extent+=$extents[$i]*$extents[$i+1]; }
              Error('Image extents exceeds maximum allowable') if $extent &&
                ($extent > (1024*$MaxImageExtent));
              $status=$image->Read("*");
              Error($status) if $#$image < 0;
              $status=$image->Write(filename=>'mpc:MagickStudio.dat');
              Error($status) if "$status";
              chop($filename);
            }
        }
      else
        {
          if ($q->param('Meta') && ($q->param('Format') ne 'Implicit'))
            {
              $filename=Untaint($q->param('Meta'));
              $q->param('SizeGeometry') =~ /(\d+)\D*(\d*)/;
              Error('Image area exceeds maximum allowable') if $1 && $2 &&
                (($1*$2) > (1024*$MaxImageArea));
              Error('Image area exceeds maximum allowable') if $1 && !$2 &&
                (($1*$1) > (1024*$MaxImageArea));
              $image=Image::Magick->new;
              $image->Set(size=>$q->param('SizeGeometry')) if
                $q->param('SizeGeometry');
              $filename=$DocumentRoot . $DocumentDirectory . '/clipboard/' .
                $q->param('SessionID') . '.mpc' if $filename eq 'clipboard:';
              $format=$q->param('Format') . ':' if $q->param('Format');
              $status=$image->Read("$format$filename$scene");
              Error($status) if $#$image < 0;
              $status=$image->Write('MagickStudio.dat');
              $status=$image->Write(filename=>'mpc:MagickStudio.dat') if
                "$status";
              Error($status) if "$status";
              $format='';
            }
          }
    }
  Error('Unable to read image file',$filename)
    unless (-f 'MagickStudio.dat') && (-s 'MagickStudio.dat');
  Error('Image size exceeds maximum allowable',$filename)
    unless (-s 'MagickStudio.dat') < (1024*$MaxFilesize);
  #
  # Read image.
  #
  $image=Image::Magick->new;
  $image->Set(font=>$DefaultFont);
  $image->Set(size=>$q->param('SizeGeometry')) if $q->param('SizeGeometry');
  @extents=$image->Ping("$format$path/MagickStudio.dat$scene");
  $extent=0;
  for ($i=0; $i < $#extents; $i+=4) { $extent+=$extents[$i]*$extents[$i+1]; }
  Error('Image extents exceeds maximum allowable') if $extent &&
    ($extent > (1024*$MaxImageExtent));
  $image=Image::Magick->new;
  $status=$image->Read("$format$path/MagickStudio.dat$scene");
  Error($status) if $#$image < 0;
  unlink('MagickStudio.dat');
  if (defined($q->param('Channel')))
    {
      my($channel);

      $channel=$q->param('Channel');
      $image->Channel($channel) unless $channel eq 'All';
    }
  $basename=$filename;
  $filename=~/^((?:.*[:\\\/])?)(.*)(\..*)/s;
  $basename=$2 if $2;
  $basename=~s/ //g;
  $q->param(-name=>'Name',-value=>$basename);
  $format=$image->Get('magick');
  $format=~tr/A-Z/a-z/;
  $q->param(-name=>'Magick',-value=>$format);
  $digest=new Digest::SHA1;
  $digest->add($q->remote_addr(),time(),$$);
  $session=$image->Get('signature') . '.' . $digest->hexdigest;
  if (defined($q->param('SessionID')))
    {
      my($clipboard, $filename);

      #
      # Paste file to clipboard.
      #
      $filename=$DocumentRoot . $DocumentDirectory . '/clipboard/' .
        $q->param('SessionID') . '.mpc';
      $clipboard=Image::Magick->new;
      $status=$clipboard->Read($filename);
      if ($#$clipboard >= 0)
        {
          $filename=$DocumentRoot . $DocumentDirectory . '/clipboard/' .
            $session;
          $status=$clipboard->Write(filename=>"$filename.mpc");
          Error($status) if "$status";
          $clipboard->Zoom($IconSize);
          $status=$clipboard->Write("$filename.gif");
          Error($status) if "$status";
        }
    }
  $q->param(-name=>'SessionID',-value=>$session);
  SaveQueryState($session,'Input');
  $q->delete('File');
  $q->delete('URL');
  #
  # Write image.
  #
  Header(GetTitle($image));
  $status=$image->Write(filename=>'MagickStudio.mpc');
  Error($status) if "$status";
  ViewForm($image);
}

#
# Get the name of a file to use as input by the MagickStudio tools.
#
sub InputForm
{
  my($action, @ChannelTypes, @InputTypes, @InterlaceTypes, $url, $version);

  #
  # Define input image formats.
  #
  @ChannelTypes=
  (
    'All',
    'Red',
    'Green',
    'Blue',
    'Matte'
  ),
  @InputTypes=
  (
    'Implicit',
    'caption',
    'cmyk',
    'fax',
    'gradient',
    'granite',
    'gray',
    'label',
    'mono',
    'netscape',
    'plasma',
    'rgb',
    'rgba',
    'text',
    'uyvy',
    'xc',
    'yuv'
  ),
  @InterlaceTypes=
  (
    'None',
    'Line',
    'Plane',
    'Partition'
  );

  #
  # Display input image form.
  #
  $url=$q->script_name();
  $q->delete('ToolType');
  $q->param(-name=>'ToolType',-value=>'Input');
  Header("ImageMagick Studio");
  print <<XXX;
To convert or manipulate your image directly from a Web page, press
<b>Browse</b> to browse and select your image file or enter the
<a href="$DocumentDirectory/URL.html">URL</a> of your image.  Next, set any of
the optional parameters below.  Finally, press <b>view</b> to continue.
<p>
XXX
  ;
  $version=Image::Magick->new()->VERSION;
  $action=$url . "?CacheID=" . $q->param('CacheID') .  "&Action=view";
  print $q->start_multipart_form(-action=>$action);
  print $q->hidden(-name=>'SessionID'), "\n";
  print "<dl>\n";
  print "<dl>\n";
  print "<dt><a href=\"$DocumentDirectory/Filename.html\">Filename</a>:\n";
  print '<dd>', $q->filefield(-name=>'File',-size=>50,maxlength=>1024), "<p>\n";
  print "<dt><a href=\"$DocumentDirectory/URL.html\">URL</a>:\n";
  print '<dd>', $q->textfield(-name=>'URL',-size=>50,-selected), "<p>\n";
  print "</dl>\n";
  print "</dl>\n";
  print 'Press to ', $q->submit(-name=>'Action',-value=>'view'),
    ' your image or ', $q->reset(-name=>'reset'), " the form.\n";
  print <<XXX;
<p>
An example
<a href="$url?File=$DocumentRoot$DocumentDirectory/images/rose.jpg&Action=view">
image</a> is available to help you get familiar with <b>ImageMagick Studio</b>,
version $version.
<p><hr width="50%" size="5"><p>
<center><h4>Privacy Notice</h4></center>
Your privacy is protected as long as you use this service in a lawful manner.
<p>
All uploaded images are temporarily stored on our local disks for processing
and they are automatically removed within a few hours.  Your images cannot
be viewed or copied by anyone other than yourself.  We have security precautions
in place to prevent others from accessing your images.
<p>
<center><h4>Liability Notice</h4></center>
By using this service, you agree not to hold ImageMagick Studio liable for any
data loss, subsequent damages, or privacy issues resulting from the use of
this service.
XXX
  ;
  print "<p><hr width=\"50%\" size=\"5\"><p>\n";
  print "<center><h4>Optional Input Attributes</h4></center>\n";
  print <<XXX;
You will rarely need to set these parameters.  The scene
specification is useful when you want to view only a few frames from a
multi-frame image.  The rest of the options are only necessary for raw
image formats such as RGB or GRAY.
<p>
<dl>
XXX
  ;
  print "<dd><table cellpadding=\"2\">\n";
  print "<tr>\n";
  print "<th><a href=\"$DocumentDirectory/Size.html\">Size</a></th>\n";
  print "<th><a href=\"$DocumentDirectory/Format.html\">Format</a></th>\n";
  print "</tr>\n";
  print "<tr>\n";
  print '<td>', $q->textfield(-name=>'SizeGeometry',-size=>20,
    -value=>'320x240'), "</td>\n";
  print '<td>', $q->popup_menu(-name=>'Format',-values=>[@InputTypes]),
    "</td>\n";
  print "</tr>\n";
  print '</table>';
  print "<dd><table cellpadding=\"2\">\n";
  print "<tr>\n";
  print "<th><a href=\"$DocumentDirectory/Meta.html\">Meta</a></th>\n";
  print "<th><a href=\"$DocumentDirectory/Interlace.html\">Interlace</a></th>\n";
  print "</tr>\n";
  print "<tr>\n";
  print '<td>', $q->textfield(-name=>'Meta',-size=>20), "</td>\n";
  print '<td>', $q->popup_menu(-name=>'Interlace',-values=>[@InterlaceTypes]),
    "</td>\n";
  print '</table>';
  print "<dd><table cellpadding=\"2\">\n";
  print "<th><a href=\"$DocumentDirectory/Scene.html\">Scene</a></th>\n";
  print "<th><a href=\"$DocumentDirectory/Channel.html\">Channel</a></th>\n";
  print "<tr>\n";
  print "</tr>\n";
  print "</tr>\n";
  print "<tr>\n";
  print '<td>', $q->textfield(-name=>'Scene',-size=>20), "</td>\n";
  print '<td>', $q->popup_menu(-name=>'Channel',-values=>[@ChannelTypes]),
    "</td>\n";
  print "</tr>\n";
  print '</table>';
  print "</dl>\n";
  print $q->endform, "\n";
  print <<XXX;
XXX
  ;
  Trailer(undef);
}

#
# Choose the appropriate Web page based on the ToolType parameter.
#
sub Mogrify
{
  my ($function, $tooltype, %Tools);

  %Tools=
  (
    'Input'=>\&InputForm,
    'View'=>\&ViewForm,
    'Identify'=>\&Identify,
    'Output'=>\&OutputForm,
    'Colormap'=>\&ColormapForm,
    'Resize'=>\&ResizeForm,
    'Transform'=>\&TransformForm,
    'Enhance'=>\&EnhanceForm,
    'Effects'=>\&EffectsForm,
    'FX'=>\&FXForm,
    'Decorate'=>\&DecorateForm,
    'Annotate'=>\&AnnotateForm,
    'Draw'=>\&DrawForm,
    'Composite'=>\&CompositeForm,
    'Comment'=>\&CommentForm
  );

  $tooltype=$q->param('ToolType');
  View() unless defined($tooltype);
  Error('Unable to view image','no path is defined') unless $q->param('Path');
  RestoreQueryState($q->param('SessionID'),$q->param('Path'),$tooltype);
  $function=$Tools{$tooltype};
  &$function if defined($function);
  Error('Request failed due to malformed query');
}

#
# Output the image in the same or differing image format.
#
sub Output
{
  use Image::Magick;

  my($basename, $format, $height, $hostname, $i, $image, %mimes, $montage,
    $path, $prefix, $size, $status, $url, $username, $value, $width, $x, $y);

  #
  # Read image.
  #
  $path=Untaint($q->param('Path'));
  chdir($path) || Error('Your image has expired',$path);
  $image=Image::Magick->new;
  $status=$image->Read("$path/MagickStudio.mpc");
  Error($status) if $#$image < 0;
  #
  # Write image.
  #
  CreateWorkDirectory(undef);
  Header(GetTitle($image));
  $status=$image->Write(filename=>'MagickStudio.mpc');
  Error($status) if "$status";
  #
  # Convert the image to the selected format.
  #
  $value=$q->param('Type');
  $image->Set(type=>$value) if $value ne 'Implicit';
  $value=$q->param('Channel');
  $image->Channel($value) unless $value eq 'All';
  $image->Set(font=>$DefaultFont);
  $value=$q->param('Compress');
  $image->Set(compression=>$value) if $value ne 'Default';
  $value=$q->param('Page');
  $image->Set(page=>$value) if length($value) > 0;
  $value=$q->param('Delay');
  $image->Set(delay=>$value) if length($value) > 0;
  $value=$q->param('Dispose');
  $image->Set(dispose=>$value);
  $value=$q->param('Loop');
  $image->Set(loop=>$value) if length($value) > 0;
  $value=$q->param('Quality');
  $image->Set(quality=>$value) if length($value) > 0;
  $value=$q->param('Interlace');
  $image->Set(interlace=>$value);
  $value=$q->param('Preview');
  $image->Set(preview=>$value);
  $image->Set(pointsize=>24);
  $image->Set(adjoin=>1);
  $image->Set(adjoin=>0) if $q->param('Option') eq 'single file';
  $image->Set(colorspace=>'CMYK') if
    $q->param('CMYK') && ($q->param('CMYK') eq 'on');
  $value=$q->param('Comment');
  $image->Comment($value) if length($value) > 0;
  if ($q->param('Option') eq 'append')
    {
      my($append_image);

      $value='True';
      $value='False' if $q->param('Stack') eq 'on';
      $image=$image->Append(stack=>$value);
      if (ref($append_image))
        {
          #
          # Replace image sequence with morph sequence.
          #
          undef $image;
          $image=$append_image;
        }
    }
  $prefix='';
  $prefix=$q->param('Option') . ':' if
    ($q->param('Option') eq 'histogram') || ($q->param('Option') eq 'preview');
  $basename=$q->param('Name');
  for ($q->Accept()) { $mimes{$_}++; }
  $format='gif';
  if ($#$image == 0)
    {
      if (($image->Get('Class') eq 'DirectClass') && !$image->Get('matte'))
        {
          $format='jpg';
          $format='png' if $mimes{'image/png'};
        }
    }
  $status=$image->Write("$basename.$format");
  Error($status) if "$status";
  $format=$q->param('Magick');
  $format=$q->param('Format') if $q->param('Format');
  $format='jpg' if $format =~ /jpeg/;
  $image->Set(magick=>$format);
  for ($i=0; $image->[$i]; $i++)
  { $image->[$i]->Set(scene=>$i); }
  if ($q->param('Option') eq 'clipboard')
    {
      my($clipboard, $filename);

      #
      # Paste file to clipboard.
      #
      $filename=$DocumentRoot . $DocumentDirectory . '/clipboard/' .
        $q->param('SessionID');
      $status=$image->Write(filename=>"$filename.mpc");
      Error($status) if "$status";
      $clipboard=$image->Clone();
      $clipboard->Zoom($IconSize);
      $status=$clipboard->Write("$filename.gif");
      Error($status) if "$status";
    }
  if ($#$image > 0)
    {
      $status=$image->Write(filename=>"$prefix$basename.$format",adjoin=>1);
      Error($status) if "$status";
      $status=$image->Write(filename=>'Animate.gif',adjoin=>1);
      Error($status) if "$status";
    }
  if ($image->Get('adjoin') || ($#$image == 0))
    {
      $status=$image->Write("$prefix$basename.$format");
      Error($status) if "$status";
    }
  else
    {
      $status=$image->Write("$prefix$basename-%d.$format");
      Error($status) if "$status";
    }
  if (($q->param('Option') eq 'histogram') ||
      ($q->param('Option') eq 'preview'))
    {
      undef @$image;
      $status=$image->Read("$basename.$format");
      Error($status) if $#$image < 0;
      $status=$image->Write("$basename.$format");
      Error($status) if "$status";
    }
  #
  # Label image.
  #
  for ($i=0; $image->[$i]; $i++)
  {
    $image->[$i]->Label($image->[$i]->Get('filename'));
    next if $i == 0;
    $image->[$i]->Label($image->[$i]->Get('scene')) if
      $image->[0]->Get('filename') eq $image->[$i]->Get('filename');
  }
  $montage=$image->Montage(background=>'#CCCCCC',borderwidth=>0,
    geometry=>'120x120+2+2>',gravity=>'Center',tile=>'4x2000',
    font=>$DefaultFont,fill=>'#600',transparent=>'#CCCCCC');
  Error($montage) if !ref($montage);
  $montage->Set(page=>'');
  $status=$montage->Write('MagickStudio.gif');
  Error($status) if "$status";
  #
  # Display images to the user.
  #
  $url=substr($q->param('Path'),length($DocumentRoot));
  print <<XXX;
<p>
Here is your converted image (or images).  Click on any image to view
or precede your click by the shift key to download it to your local
area or press <b>upload</b> to transfer the image to a remote site.
<p>
<center>
XXX
  ;
  if ($#$image > 0)
    {
      #
      # Display animated image.
      #
      ($width,$height)=$image->Get('width','height');
      print <<XXX;
<a href="$url/$basename.$format"> <img alt="$basename.$format"
  src="$url/Animate.gif" width=$width height=$height border="0"></a>
<p><hr width="50%" size="5"><p>
XXX
  ;
    }
  ($width,$height)=$montage->Get('width','height');
  print <<XXX;
<img ismap usemap=#Montage src="$url/MagickStudio.gif" width=$width height=$height border="0">
<map name=Montage>
XXX
  ;
  $montage->Get('montage')=~/(\d+)x(\d+)\+(\d+)\+(\d+)/;
  $width=$1;
  $height=$2;
  $x=$3;
  $y=$4;
  for (split(/\n/,$montage->Get('directory')))
  {
    print "  <area href=\"$url/$_\"", " shape=rect coords=$x,$y,",
      $x+$width-1, ',', $y+$height-1, ">\n";
    $x+=$width;
    if ($x >= $montage->Get('width'))
      {
        $x=0;
        $y+=$height;
      }
  }
  $hostname=$q->server_name();
  print <<XXX;
</map>
<p>
<a href="http://transload.starblvd.net/cgi-bin/transload?from=http://$hostname$url/$basename.$format&provider=ImageMagick+Studio">
<font face="Arial,Helvetica" size=-1>Transload this picture to your
website!</font>
</a>
<p>
<br>
<br>
</center>
XXX
  ;
  print "* <i>this image was saved to the ImageMagick Studio clipboard.</i><p>"
    if $q->param('Option') eq 'clipboard';
  #
  # Image upload form.
  #
  RestoreQueryState($q->param('SessionID'),$q->param('Path'),'Upload');
  print $q->startform;
  print $q->hidden(-name=>'CacheID'), "\n";
  print $q->hidden(-name=>'SessionID'), "\n";
  print $q->hidden(-name=>'Path'), "\n";
  print $q->hidden(-name=>'ToolType'), "\n";
  print $q->hidden(-name=>'Name'), "\n";
  print $q->hidden(-name=>'Magick'), "\n";
  print $q->hidden(-name=>'Action',-value=>'upload'), "\n";
  print 'Press to ', $q->submit(-name=>'Action',-value=>'upload'),
    ' your image to a remote site or ', $q->reset(-name=>'reset'),
    " the form.\n";
  print "<p><hr width=\"50%\" size=\"5\"><p>\n";
  print "<center><h4>Optional Upload Attributes</h4></center>\n";
  print "</center>\n";
  print "<dl>\n";
  $hostname=GetHostname($q->remote_host());
  print "<dt><a href=\"$DocumentDirectory/Upload.html\">FTP server name</a>:\n";
  print '<dd>', $q->textfield(-name=>'Hostname',-size=>50,-value=>$hostname),
    "<p>\n";
  print "<dt><a href=\"$DocumentDirectory/Upload.html\">Account name</a>:\n";
  $username='anonymous';
  $username=$q->remote_user() if $q->remote_user();
  print '<dd>', $q->textfield(-name=>'Username',-size=>20,-value=>$username),
    "<p>\n";
  print "<dt><a href=\"$DocumentDirectory/Upload.html\">Account password</a>:\n";
  print '<dd>', $q->password_field(-name=>'Password',-size=>20), "<p>\n";
  print "<dt><a href=\"$DocumentDirectory/Upload.html\">Upload directory</a>:\n";
  print '<dd>', $q->textfield(-name=>'Directory',-size=>50), "<p>\n";
  print "<dt><a href=\"$DocumentDirectory/Upload.html\">Filename</a>:\n";
  print '<dd>', $q->textfield(-name=>'Filename',-size=>50,
    -value=>"$basename.$format"), "<p>\n";
  print "</dl>\n";
  print $q->endform, "\n";
  Trailer(undef);
}

#
# Output image form.
#

sub SelectFormats
{
  my($image) = @_;

  return undef unless ($image->QueryFormat($_))[4];
  $_;
}

sub OutputForm
{
  my(@ChannelTypes, @CompressTypes, @DisposeTypes, $format, @formats, $image,
    @ImageTypes, @InterlaceTypes, @OptionTypes, $path, @PreviewTypes, $status);

  @ChannelTypes=
  (
    'All',
    'Red',
    'Green',
    'Blue',
    'Matte'
  ),
  @CompressTypes=
  (
    'Default',
    'None',
    'BZip',
    'Fax',
    'JPEG',
    'LZW',
    'RLE',
    'Zip'
  ),
  @DisposeTypes=
  (
    'Undefined',
    'None',
    'Background',
    'Previous'
  ),
  @ImageTypes=
  (
    'Implicit',
    'Bilevel',
    'Grayscale',
    'GrayscaleMatte',
    'Palette',
    'PaletteMatte',
    'Truecolor',
    'TruecolorMatte',
    'ColorSeparation',
    'ColorSeparationMatte'
  ),
  @InterlaceTypes=
  (
    'None',
    'Line',
    'Plane',
    'Partition'
  ),
  @OptionTypes=
  [
    'single file',
    'append',
    'preview',
    'histogram',
    'clipboard',
    'multi-frame file'
  ],
  @PreviewTypes=
  (
    'Gamma',
    'Rotate',
    'Shear',
    'Roll',
    'Hue',
    'Saturation',
    'Brightness',
    'Spiff',
    'Dull',
    'Grayscale',
    'Quantize',
    'Despeckle',
    'ReduceNoise',
    'AddNoise',
    'Sharpen',
    'Blur',
    'Threshold',
    'EdgeDetect',
    'Spread',
    'Solarize',
    'Shade',
    'Raise',
    'Segment',
    'Swirl',
    'Implode',
    'Wave',
    'OilPaint',
    'CharcoalDrawing',
    'JPEG'
  );

  #
  # Read image.
  #
  $path=Untaint($q->param('Path'));
  chdir($path) || Error('Your image has expired',$path);
  $image=Image::Magick->new;
  $status=$image->Read("$path/MagickStudio.mpc");
  Error($status) if $#$image < 0;
  #
  # Display Output form.
  #
  Header(GetTitle($image));
  print <<XXX;
Choose an <a href="$DocumentDirectory/Output.html">output</a> image format and
set any optional image attributes below.  Some attributes are only relevant to
specific output formats.  Next, press <b>output</b> to convert your image to
the selected format.  The image is converted and you are given an opportunity
to download it to your local area.
<p>
XXX
  ;
  print $q->startform;
  print $q->hidden(-name=>'CacheID'), "\n";
  print $q->hidden(-name=>'SessionID'), "\n";
  print $q->hidden(-name=>'Path'), "\n";
  print $q->hidden(-name=>'ToolType'), "\n";
  print $q->hidden(-name=>'Name'), "\n";
  print $q->hidden(-name=>'Magick'), "\n";
  print $q->hidden(-name=>'Action',-value=>'output'), "\n";
  print "<dl>\n";
  print "<dl>\n";
  print "<dt><a href=\"$DocumentDirectory/Format.html\">Format</a>:\n";
  $format=$q->param('Magick');
  @formats=grep(SelectFormats($image),$image->QueryFormat());
  print '<dd>', $q->scrolling_list(-name=>'Format',-values=>[@formats],
    -size=>5,-default=>$format), "<p>\n";
  print "<dt><a href=\"$DocumentDirectory/Storage.html\">Storage type</a>:\n";
  print '<dd>', $q->radio_group(-name=>'Option',-values=>@OptionTypes,
    -columns=>3,-default=>'multi-frame file'), "<p>\n";
  print "</dl>\n";
  print "</dl>\n";
  print 'Press to ', $q->submit(-name=>'Action',-value=>'output'),
    ' your image or ', $q->reset(-name=>'reset'), " the form.\n";
  print "<p><hr width=\"50%\" size=\"5\"><p>\n";
  print "<center><h4>Optional Output Attributes</h4></center>\n";
  print "<dl>\n";
  print "<dd><table cellpadding=\"2\">\n";
  print "<tr>\n";
  print "<th>Image Type</th>\n";
  print "<th>Compress</th>\n";
  print "<th><a href=\"$DocumentDirectory/Channel.html\">Channel</a></th>\n";
  print "</tr>\n";
  print "<tr>\n";
  print '<td>', $q->popup_menu(-name=>'Type',-values=>[@ImageTypes]), "</td>\n";
  print '<td>', $q->popup_menu(-name=>'Compress',-values=>[@CompressTypes]),
    "</td>\n";
  print '<td>', $q->popup_menu(-name=>'Channel',-values=>[@ChannelTypes]),
    "</td>\n";
  print "</tr>\n";
  print "</table>\n";
  print "<dd><table cellpadding=\"2\">\n";
  print "<tr>\n";
  print "<th>Preview</th>\n";
  print "<th>Dispose</th>\n";
  print "<th><a href=\"$DocumentDirectory/Interlace.html\">Interlace</a></th>\n";
  print "</tr>\n";
  print "<tr>\n";
  print '<td>', $q->popup_menu(-name=>'Preview',-values=>[@PreviewTypes]),
    "</td>\n";
  print '<td>', $q->popup_menu(-name=>'Dispose',-values=>[@DisposeTypes]),
    "</td>\n";
  print '<td>', $q->popup_menu(-name=>'Interlace',-values=>[@InterlaceTypes]),
    "</td>\n";
  print "</tr>\n";
  print "</table>\n";
  print "<dd><table cellpadding=\"2\">\n";
  print "<tr>\n";
  print "<th><a href=\"$DocumentDirectory/Delay.html\">Delay</a></th>\n";
  print "<th><a href=\"$DocumentDirectory/Loop.html\">Loop</a></th>\n";
  print "<th><a href=\"$DocumentDirectory/Quality.html\">Quality</a></th>\n";
  print "</tr>\n";
  print "<tr>\n";
  print '<td>', $q->textfield(-name=>'Delay',-size=>15), "</td>\n";
  print '<td>', $q->textfield(-name=>'Loop',-size=>15,-value=>'0'), "</td>\n";
  print '<td>', $q->textfield(-name=>'Quality',-value=>75,-size=>15), "</td>\n";
  print "</tr>\n";
  print "</table><p>\n";
  print "<dt><a href=\"$DocumentDirectory/Page.html\">Page Geometry</a><p>\n";
  print '<dd>', $q->textfield(-name=>'Page',-size=>15), "<p>\n";
  print "<dt>Comment:\n";
  print '<dd>', $q->textarea(-name=>'Comment',-columns=>50,-rows=>3), "<p>\n";
  print "<dt> Miscellaneous options:\n";
  print '<dd> ', $q->checkbox(-name=>'CMYK',
    -label=>' save image as CMYK pixels (JPEG, TIFF, PS, PDF, PSD)'),"\n";
  print '<dd> ', $q->checkbox(-name=>'Stack',
    -label=>' stack images left-to-right (when storage type is append)'),
    "<p>\n";
  print "</dl>\n";
  print $q->endform, "\n";
  print <<XXX;
XXX
  ;
  Trailer(undef);
}

#
# Resize image.
#
sub Resize
{
  use Image::Magick;

  my($blur, $filter, $geometry, $image, $montage, $path, $status, $value);

  #
  # Read image.
  #
  $path=Untaint($q->param('Path'));
  chdir($path) || Error('Your image has expired',$path);
  $image=Image::Magick->new;
  $status=$image->Read("$path/MagickStudio.mpc");
  Error($status) if $#$image < 0;
  #
  # Resize image.
  #
  $image->Set(gravity=>$q->param('Gravity'));
  $q->param('Geometry') =~ /(\d+)\D*(\d*)/;
  Error('Image area exceeds maximum allowable') if $1 && $2 &&
    (($1*$2) > (1024*$MaxImageArea));
  Error('Image area exceeds maximum allowable') if $1 && !$2 &&
    (($1*$1) > (1024*$MaxImageArea));
  $geometry='100%';
  $geometry=$q->param('Geometry') if $q->param('Geometry');
  $filter='Undefined';
  $filter=$q->param('FilterType') if $q->param('FilterType');
  $blur='1.0';
  $blur=$q->param('BlurFactor') if $q->param('BlurFactor');
  $image->Resize(geometry=>$geometry,filter=>$filter,blur=>$blur) if
    $q->param('Algorithm') eq 'resize *';
  $image->Scale($geometry) if $q->param('Algorithm') eq 'scale *';
  $image->Sample($geometry) if $q->param('Algorithm') eq 'sample *';
  $image->Magnify() if $q->param('Algorithm') eq 'double size';
  $image->Minify() if $q->param('Algorithm') eq 'half size';
  $image->Resample(density=>$geometry,filter=>$filter,blur=>$blur) if
    $q->param('Algorithm') eq 'resample *';
  if ($q->param('Algorithm') eq 'thumbnail *')
    {
      my($mode, $tile_geometry);

      #
      # Thumbnail.
      #
      $mode='unframe';
      $mode=$q->param('Mode') if $q->param('Mode');
      $tile_geometry='6x4';
      $tile_geometry=$q->param('TileGeometry') if $q->param('TileGeometry');
      $montage=$image->Montage(geometry=>$geometry,tile=>$tile_geometry,
        mode=>$mode,background=>'#CCCCCC',font=>$DefaultFont,fill=>'#600',
        gravity=>'Center',transparent=>'#CCCCCC');
      $image=$montage;
    }
  #
  # Write image.
  #
  CreateWorkDirectory(1);
  Header(GetTitle($image));
  $status=$image->Write(filename=>'MagickStudio.mpc');
  Error($status) if "$status";
  ViewForm($image);
}

#
# Resize image form.
#
sub ResizeForm
{
  use Image::Magick;

  my(@FilterTypes, @GravityTypes, $height, $image, @ModeTypes, @OptionTypes,
    $path, $width);

  @FilterTypes=
  (
    'Undefined',
    'Point',
    'Box',
    'Triangle',
    'Hermite',
    'Hanning',
    'Hamming',
    'Blackman',
    'Gaussian',
    'Mitchell',
    'Lanczos',
    'Quadratic',
    'Cubic',
    'Catrom',
    'Bessel',
    'Sinc'
  ),
  @GravityTypes=
  (
    'NorthWest',
    'North',
    'NorthEast',
    'West',
    'Center',
    'East',
    'SouthWest',
    'South',
    'SouthEast'
  ),
  @ModeTypes=
  [
    'unframe',
    'frame',
    'concatenate'
  ],
  @OptionTypes=
  [
    'resize *',
    'half size',
    'sample *',
    'double size',
    'scale *',
    'resample *',
    'thumbnail *'
  ];

  #
  # Read image.
  #
  $path=Untaint($q->param('Path'));
  chdir($path) || Error('Your image has expired',$path);
  $image=Image::Magick->new;
  ($width,$height)=$image->Ping("$path/MagickStudio.mpc");
  $q->delete('Geometry');
  #
  # Display Resize form.
  #
  Header(GetTitle($image));
  print <<XXX;
To <a href="$DocumentDirectory/Resize.html">resize</a> your image, specify the
desired size and scaling method.  Note, only methods denoted with an asterisk
require a parameter value.  Next, press <b>resize</b> to continue.
XXX
  ;
  print $q->startform;
  print $q->hidden(-name=>'CacheID'), "\n";
  print $q->hidden(-name=>'SessionID'), "\n";
  print $q->hidden(-name=>'Path'), "\n";
  print $q->hidden(-name=>'ToolType'), "\n";
  print $q->hidden(-name=>'Name'), "\n";
  print $q->hidden(-name=>'Magick'), "\n";
  print $q->hidden(-name=>'Action',-value=>'resize'), "\n";
  print "<dl>\n";
  print "<dl>\n";
  print "<dt>Image size:\n";
  print '<dd>', $q->textfield(-name=>'Geometry',-size=>20,
    -value=>"$width" . 'x' . "$height"), "<p>\n";
  print "<dt>Choose from these scaling methods</a>:\n";
  print '<dd>', $q->radio_group(-name=>'Algorithm',-values=>@OptionTypes,
    -columns=>3), "<p>\n";
  print "</dl>\n";
  print "</dl>\n";
  print 'Press to ', $q->submit(-name=>'Action',-value=>'resize'),
    ' your image or ', $q->reset(-name=>'reset'), " the form.\n";
  print "<p><hr width=\"50%\" size=\"5\"><p>\n";
  print "<center><h4>Optional Resize Attributes</h4></center>\n";
  print "<dl>\n";
  print "<dd><table cellpadding=\"2\">\n";
  print "<tr>\n";
  print "<th>Blur</th>\n";
  print "<th>Filter</th>\n";
  print "</tr>\n";
  print '<td>', $q->textfield(-name=>'BlurFactor',-size=>15,-value=>'1.0'),
    "</td>\n";
  print '<td>', $q->popup_menu(-name=>'Primitive',-values=>[@FilterTypes]),
    "</td>\n";
  print "</tr>\n";
  print '</table><p>';
  print "<dd><table cellpadding=\"2\">\n";
  print "<tr>\n";
  print "<th>Thumbnail</th>\n";
  print "<th>Gravity</th>\n";
  print "</tr>\n";
  print "<tr>\n";
  print '<td>', $q->textfield(-name=>'TileGeometry',-size=>15,-value=>'6x4'),
    "</td>\n";
  print '<td>', $q->popup_menu(-name=>'Gravity',-values=>[@GravityTypes]),
    "</td>\n";
  print "</tr>\n";
  print '</table><p>';
  print "<dt>Framing mode (thumbnail):\n";
  print '<dd>', $q->radio_group(-name=>'Mode',-values=>@ModeTypes,-columns=>3),
    "<p>\n";
  print "</dl>\n";
  print $q->endform, "\n";
  print <<XXX;
XXX
  ;
  Trailer(1);
}

#
# Restore query state from a file.
#
sub RestoreQueryState
{
  no strict 'subs';

  my($session, $path, $tooltype) = @_;

  my($filename);

  local(*DATA);

  $filename=Untaint($DocumentRoot . $DocumentDirectory .
    "/session_info/$session.$tooltype");
  open(DATA,$filename) || return;
  $q=CGI->new(DATA);
  close(DATA);
  $q->param(-name=>'Path',-value=>$path);
}

#
# Save query state to a file.
#
sub SaveQueryState
{
  no strict 'subs';

  my($session, $tooltype) = @_;

  my($filename, $path, $remote_host);

  local(*DATA);

  $path=$DocumentRoot . $DocumentDirectory . "/session_info";
  $filename=Untaint("$path/$session.$tooltype");
  open(DATA,">$filename") || Error('Unable to view image',$filename);
  $q->save(DATA);
  print DATA "image locator: ", $q->param('File'), "\n" if $q->param('File');
  print DATA "image locator: ", $q->param('URL'), "\n" if $q->param('URL');
  $remote_host='localhost';
  $remote_host=$q->remote_host() if $q->remote_host();
  print DATA "remote host: ", GetHostname($remote_host), "\n";
  print DATA "remote addr: ", $q->remote_addr(), "\n" if $q->remote_addr();
  print DATA "remote ident: ", $q->remote_ident(), "\n" if $q->remote_ident();
  print DATA "remote user: ", $q->remote_user(), "\n" if $q->remote_user();
  print DATA "user name: ", $q->user_name(), "\n" if  $q->user_name();
  print DATA "user agent: ", $q->user_agent(), "\n" if $q->user_agent();
  print DATA "time stamp: ", scalar(localtime), "\n";
  close(DATA);
}

#
# Print the standard HTML trailer.
#
sub Trailer
{
  my($display) = @_;

  my($home, $load_average, $url);

  print <<XXX;
<p><hr><p>
XXX
  ;
  if ($display)
    {
      my($filename, $height, $image, $path, $url, $width);

      #
      # Display image and clipboard icon.
      #
      $image=Image::Magick->new;
      $path="";
      $path=Untaint($q->param('Path')) if $q->param('Path');
      $filename="$path/MagickStudio.gif";
      if (-e $filename)
        {
          ($width,$height)=$image->Ping("$filename" . '[0]');
          $url=substr($path,length($DocumentRoot));
          print "<img alt=\"image icon\" src=\"$url/MagickStudio.gif\" ",
            "border=\"0\" width=\"$width\" height=\"$height\" align=\"right\"></a>\n";
        }
      $filename=$DocumentRoot . $DocumentDirectory . '/clipboard/' .
        $q->param('SessionID') . '.gif';
      if (-e $filename)
        {
          ($width,$height)=$image->Ping("$filename" . '[0]');
          print "<img alt=\"clipboard icon\" src=\"$DocumentDirectory/",
            "clipboard/", $q->param('SessionID'), ".gif\" border=\"0\" ",
            "width=\"$width\" height=\"$height\" align=\"right\"></a>\n";
        }
    }
  #
  # Display home and comment icon.
  #
  $url=$q->script_name;
  $load_average=GetLoadAverage();
  $home="$DocumentDirectory/images/home-go.png";
  $home="$DocumentDirectory/images/home-caution.png"
    if $load_average >= ($LoadAverageThreshold/3);
  $home="$DocumentDirectory/images/home-stop.png"
    if $load_average >= (2*$LoadAverageThreshold/3);
  print <<XXX;
<a href="$url"> <img alt="home" src=$home width="40" height="40" border="0"></a>
<a href="$url?SessionID=null&Path=null&Action=mogrify&ToolType=Comment">
<img alt="comment" src="$DocumentDirectory/images/mail.png" width="40" height="40" border="0"></a>
</td></tr></table>
XXX
  ;
  if ($Debug)
    {
      my($name);

      #
      # Display debugging information for an administrator.
      #
      print "<p><hr width=\"50%\" size=\"5\"><p>\n";
      print "<center><h3>CGI State Information</h3></center>\n";
      print "Script:\n";
      print '<dd>', $0, "<p>\n";
      print "Action:\n";
      print '<dd>', $action, "<p>\n";
      print "Time:\n";
      print '<dd>', time-$timer, "s<p>\n";
      print "Query state:\n";
      print '<dd>', $q->query_string, "<p>\n";
      print $q->Dump;
      print "<p>\n";
      print "Environment state:\n";
      print "<ul>\n";
      print "<li><b>accept</b>:", $q->Accept(), "<br>\n";
      print "<li><b>auth type</b>:", $q->auth_type(), "<br>\n";
      print "<li><b>raw cookie</b>:", $q->raw_cookie(), "<br>\n";
      print "<li><b>path info</b>:", $q->path_info(), "<br>\n";
      print "<li><b>path translated</b>:", $q->path_translated(), "<br>\n";
      print "<li><b>referer</b>:", $q->referer(), "<br>\n";
      print "<li><b>remote addr</b>:", $q->remote_addr(), "<br>\n";
      print "<li><b>remote ident</b>:", $q->remote_ident(), "<br>\n";
      print "<li><b>remote host</b>:", GetHostname($q->remote_host()), "<br>\n";
      print "<li><b>remote user</b>:", $q->remote_user(), "<br>\n";
      print "<li><b>request method</b>:", $q->request_method(), "<br>\n";
      print "<li><b>script name</b>:", $q->script_name(), "<br>\n";
      print "<li><b>server name</b>:", $q->server_name(), "<br>\n";
      print "<li><b>server software</b>:", $q->server_software(), "<br>\n";
      print "<li><b>server port</b>:", $q->server_port(), "<br>\n";
      print "<li><b>temporary directory</b>:", $ENV{TMPDIR}, "<br>\n";
      print "<li><b>user agent</b>:", $q->user_agent(), "<br>\n";
      print "<li><b>user name</b>:", $q->user_name(), "<br>\n";
      print "<li><b>virtual host</b>:", $q->virtual_host(), "<br>\n";
      print "<li><b>environment</b>:", $q->http(), "<br>\n";
      print "<li><b>time stamp</b>:", scalar(localtime), "<br>\n";
      print "</ul>\n";
      print "<p><hr width=\"50%\" size=\"5\"><p>\n";
    }
  print $q->end_html;
  exit;
}

#
# Transform image.
#
sub Transform
{
  use Image::Magick;

  my($color, $image, $parameter, $path, $status);

  #
  # Read image.
  #
  $path=Untaint($q->param('Path'));
  chdir($path) || Error('Your image has expired',$path);
  $image=Image::Magick->new;
  $status=$image->Read("$path/MagickStudio.mpc");
  Error($status) if $#$image < 0;
  #
  # Transform image.
  #
  $parameter=$q->param('Parameter');
  $image->Set(gravity=>$q->param('Gravity'));
  $image->AffineTransform([split(/,/,$parameter)]) if
    $q->param('Option') eq 'affine *';
  $image->Trim() if $q->param('Option') eq 'trim';
  $image->Crop("$parameter") if $q->param('Option') eq 'crop *';
  $image->Chop("$parameter") if $q->param('Option') eq 'chop *';
  $image->Coalesce() if $q->param('Option') eq 'coalesce';
  $image->Deconstruct() if $q->param('Option') eq 'deconstruct';
  $image->Flatten() if $q->param('Option') eq 'flatten';
  $image->Flop() if $q->param('Option') eq 'flop';
  $image->Flip() if $q->param('Option') eq 'flip';
  $image->Rotate(90) if $q->param('Option') eq 'rotate right';
  $image->Rotate(-90) if $q->param('Option') eq 'rotate left';
  $color='none';
  $color=$q->param('BackgroundColor') if $q->param('BackgroundColor');
  $image->Rotate(degrees=>$parameter,color=>$color)
    if $q->param('Option') eq 'rotate *';
  $image->Shave("$parameter") if $q->param('Option') eq 'shave *';
  $image->Shear("$parameter") if $q->param('Option') eq 'shear *';
  $image->Roll("$parameter") if $q->param('Option') eq 'roll *';
  #
  # Write image.
  #
  CreateWorkDirectory(1);
  Header(GetTitle($image));
  $status=$image->Write(filename=>'MagickStudio.mpc');
  Error($status) if "$status";
  ViewForm($image);
}

#
# Transform image form.
#
sub TransformForm
{
  my(@GravityTypes, @OptionTypes);

  @GravityTypes=
  (
    'NorthWest',
    'North',
    'NorthEast',
    'West',
    'Center',
    'East',
    'SouthWest',
    'South',
    'SouthEast'
  );

  @OptionTypes=
  [
    'affine *',
    'chop *',
    'coalesce',
    'crop *',
    'deconstruct',
    'flatten',
    'flip',
    'flop',
    'roll *',
    'rotate *',
    'rotate left',
    'rotate right',
    'shave *',
    'shear *',
    'trim'
  ];

  #
  # Display Transform form.
  #
  Header(GetTitle(undef));
  print <<XXX;
To <a href="$DocumentDirectory/Transform.html">transform</a> your image, enter
your transform parameter and method.  Note, only methods denoted with an
asterisk require a parameter value.  Next, press <b>transform</b> to continue.
XXX
  ;
  print $q->startform;
  print $q->hidden(-name=>'CacheID'), "\n";
  print $q->hidden(-name=>'SessionID'), "\n";
  print $q->hidden(-name=>'Path'), "\n";
  print $q->hidden(-name=>'ToolType'), "\n";
  print $q->hidden(-name=>'Name'), "\n";
  print $q->hidden(-name=>'Magick'), "\n";
  print $q->hidden(-name=>'Action',-value=>'resize'), "\n";
  print "<dl>\n";
  print "<dl>\n";
  print "<dt>Parameter:\n";
  print '<dd>', $q->textfield(-name=>'Parameter',-size=>20), "<p>\n";
  print "<dt>Choose from these transforms</a>:\n";
  print '<dd>', $q->radio_group(-name=>'Option',-values=>@OptionTypes,
    -columns=>3,-default=>'trim'), "<p>\n";
  print "</dl>\n";
  print "</dl>\n";
  print 'Press to ', $q->submit(-name=>'Action',-value=>'transform'),
    ' your image or ', $q->reset(-name=>'reset'), " the form.\n";
  print "<p><hr width=\"50%\" size=\"5\"><p>\n";
  print "<center>\n";
  print "<h4>Optional Transform Attributes</h4>\n";
  print "</center>\n";
  print "<dl>\n";
  print "<dd><table cellpadding=\"2\">\n";
  print "<tr>\n";
  print "<th><a href=\"$DocumentDirectory/Color.html\">Background</a></th>\n";
  print "<th>Gravity</th>\n";
  print "</tr>\n";
  print "<tr>\n";
  print '<td>', $q->textfield(-name=>'BackgroundColor',-value=>'none',
    -size=>15), "</td>\n";
  print '<td>', $q->popup_menu(-name=>'Gravity',-values=>[@GravityTypes]),
    "</td>\n";
  print "</tr>\n";
  print '</table><p>';
  print "</dl>\n";
  print $q->endform, "\n";
  print <<XXX;
XXX
  ;
  Trailer(1);
}

#
# Untaint Perl variable.
#
sub Untaint
{
  my($filename) = @_;

  Error('Your filename contains prohibited characters',@_) if
    ($filename =~ /[\<\>\"\'\%\;\)\(\&\+]/);
  $filename;
}

#
# Upload the image to a remote FTP server.
#
sub Upload
{
  use Image::Magick;
  use URI::URL;
  use LWP::UserAgent;
  use LWP::MediaTypes qw(media_suffix);

  my($basename, $directory, $content, $filename, $format, $hostname,
     $image, $password, $path, $request, $response, $status, $url,
     $user_agent, $username);

  #
  # Read image.
  #
  SaveQueryState($q->param('SessionID'),'Upload');
  $path=Untaint($q->param('Path'));
  chdir($path) || Error('Your image has expired',$path);
  $image=Image::Magick->new;
  $status=$image->Read("$path/MagickStudio.mpc");
  $basename=$q->param('Name');
  $format=$q->param('Magick');
  Error($status) if $#$image < 0;
  #
  # Construct FTP URL.
  #
  $action='output';
  $hostname=GetHostname($q->remote_host());
  $hostname=$q->param('Hostname') if $q->param('Hostname');
  $username='anonymous';
  $username=$q->param('Username') if $q->param('Username');
  $password="$username\@$hostname";
  $password=$q->param('Password') if $q->param('Password');
  $directory=$q->param('Directory') if $q->param('Directory');
  $filename="$basename.$format";
  $filename=$q->param('Filename') if $q->param('Filename');
  $url="ftp://$hostname/$filename";
  $url="ftp://$hostname/$directory/$filename" if $q->param('Directory');
  $q->delete('Password');
  $q->delete('Filename');
  SaveQueryState($q->param('SessionID'),'Upload');
  #
  # Upload image.
  #
  $user_agent=new LWP::UserAgent;
  $user_agent->agent('MagickStudio/1.0 ' . $user_agent->agent);
  $user_agent->env_proxy if $ENV{'ftp_proxy'};
  $user_agent->timeout($Timeout);
  $request=HTTP::Request->new(PUT=>$url);
  $request->header('Content-Type','C');
  $request->authorization_basic($username,$password);
  $request->proxy_authorization_basic($username,$password);
  $image->Set(magick=>$format);
  $request->content($image->ImageToBlob());
  $response=$user_agent->request($request);
  Error($url,$response->error_as_HTML) unless $response->is_success;
  Warning('Image uploaded',$url);
}

#
# View image.
#
sub View
{
  use Image::Magick;

  my($basename, $bordercolor, $extent, @extents, $fill, $i, $image, $method,
    $path, $points, $primitive, $status);

  #
  # Read image.
  #
  $path=Untaint($q->param('Path'));
  chdir($path) || Error('Your image has expired',$path);
  $image=Image::Magick->new;
  @extents=$image->Ping("$path/MagickStudio.mpc");
  $extent=0;
  for ($i=0; $i < $#extents; $i+=4) { $extent+=$extents[$i]*$extents[$i+1]; }
  Error('Image extents exceeds maximum allowable') if $extent &&
    ($extent > (1024*$MaxImageExtent));
  $status=$image->Read("$path/MagickStudio.mpc");
  Error($status) if $#$image < 0;
  if ($q->param('Primitive') ne 'None')
    {
      #
      # Paint image.
      #
      $image->Set(fuzz=>$q->param('Fuzz')) if $q->param('Fuzz');
      $primitive=$q->param('Primitive');
      $basename=$q->param('Name');
      $points=$q->param("$basename.x") . ',' . $q->param("$basename.y");
      $fill='none';
      $fill=$q->param('Color') if $q->param('Color');
      if ($fill eq '@clipboard')
        {
          $fill='@' . $DocumentRoot . $DocumentDirectory . '/clipboard/' .
            $q->param('SessionID') . '.mpc';
        }
      $bordercolor='none';
      $bordercolor=$q->param('BorderColor') if $q->param('BorderColor');
      $method=$q->param('Method');
      $image->Draw(primitive=>$primitive,points=>$points,fill=>$fill,
        bordercolor=>$bordercolor,method=>$method);
    }
  #
  # Write image.
  #
  CreateWorkDirectory(1);
  Header(GetTitle($image));
  $status=$image->Write(filename=>'MagickStudio.mpc');
  Error($status) if "$status";
  ViewForm($image);
}

#
# Display View form.
#
sub ViewForm
{
  use Image::Magick;

  my($image) = @_;

  my($basename, $filesize, $format, $height, $icon, @MethodTypes, %mimes, $path,
    @PrimitiveTypes, $status, $url, $width);

  @MethodTypes=
  (
    'Floodfill',
    'Point',
    'Replace',
    'FillToBorder',
    'Reset'
  ),
  @PrimitiveTypes=
  (
    'None',
    'Color',
    'Matte'
  );

  #
  # Ensure image resources are within bounds.
  #
  $path=Untaint($q->param('Path'));
  chdir($path) || Error('Your image has expired',$path);
  if (!$image)
    {
      #
      # Read image.
      #
      $image=Image::Magick->new;
      $status=$image->Read("$path/MagickStudio.mpc");
      Error($status) if $#$image < 0;
      Header(GetTitle($image));
    }
  #
  # Convert the image to viewable format.
  #
  $icon=$image->Clone();
  $icon->Zoom($IconSize);
  $status=$icon->Write(filename=>'MagickStudio.gif');
  Error($status) if "$status";
  $basename=$q->param('Name');
  for ($q->Accept()) { $mimes{$_}++; }
  $format='gif';
  if ($#$image > 0)
    { $image->Set(loop=>0,delay=>800) if $image->Get('iterations') == 1; }
  else
    {
      if (($image->Get('Class') eq 'DirectClass') && !$image->Get('matte'))
        {
          $format='jpg';
          $format='png' if $mimes{'image/png'};
        }
    }
  $status=$image->Write("$basename.$format");
  Error($status) if "$status";
  #
  # Display View form.
  #
  ($width,$height)=$image->Get('width','height');
  $url=substr($path,length($DocumentRoot));
  print <<XXX;
Here is your image.  Click on a tab above to interactively resize,
rotate, sharpen, color reduce, or add special effects to your image and
save the completed work in the same or differing image format.  Press
<b>Back</b> to undo your last image transformation.  For more
information, see
<a href="http://www.imagemagick.org/">ImageMagick</a>.
<p>
You can optionally
<a href="$DocumentDirectory/Paint.html">paint</a> on your image.  Set any
optional attributes below and click on the appropriate location within your
image.
<p><hr width="50%" size="5"> <p>
XXX
  ;
  print $q->startform;
  print $q->hidden(-name=>'CacheID'), "\n";
  print $q->hidden(-name=>'SessionID'), "\n";
  print $q->hidden(-name=>'Path'), "\n";
  print $q->hidden(-name=>'ToolType'), "\n";
  print $q->hidden(-name=>'Name'), "\n";
  print $q->hidden(-name=>'Magick'), "\n";
  print $q->hidden(-name=>'Action',-value=>'view'), "\n";
  print "<center>\n";
  print $q->image_button(-name=>$basename,-src=>"$url/$basename.$format",
    -width=>$width,-height=>$height,-style=>"cursor:crosshair",-cursor),
    "<p>\n";
  if (defined($q->param("$basename.x")) || defined($q->param("$basename.y")))
    {
      my($color, $x, $y);

      $x=$q->param("$basename.x");
      $y=$q->param("$basename.y");
      $color=$image->Get("pixel[$x,$y]");
      print "<pre>$x,$y: rgba($color) \t",
        $image->QueryColorname("rgba($color)"), "</pre>\n" if $color;
    }
  print "<p><hr width=\"50%\" size=\"5\"><p>\n";
  print "<h4>Optional Paint Attributes</h4>\n";
  print "</center>\n";
  print "<dl>\n";
  print "<dd><table cellpadding=\"2\">\n";
  print "<tr>\n";
  print "<th><a href=\"$DocumentDirectory/Fuzz.html\">Fuzz</a></th>\n";
  print "<th><a href=\"$DocumentDirectory/Paint.html\">Method</a></th>\n";
  print "<th><a href=\"$DocumentDirectory/Paint.html\">Paint Type</a></th>\n";
  print "</tr>\n";
  print "<tr>\n";
  print '<td>', $q->textfield(-name=>'Fuzz',-size=>15,-value=>'0'), "</td>\n";
  print '<td>', $q->popup_menu(-name=>'Method',-values=>[@MethodTypes]),
    "</td>\n";
  print '<td>', $q->popup_menu(-name=>'Primitive',-values=>[@PrimitiveTypes]),
    "</td>\n";
  print "</tr>\n";
  print '</table><p>';
  print "<dd><table cellpadding=\"2\">\n";
  print "<tr>\n";
  print "<th><a href=\"$DocumentDirectory/Color.html\">Fill Color</a></th>\n";
  print "<th><a href=\"$DocumentDirectory/Color.html\">Border Color</a></th>\n";
  print "</tr>\n";
  print "<tr>\n";
  print '<td>', $q->textfield(-name=>'Color',-value=>'white',-size=>15),
    "</td>\n";
  print '<td>', $q->textfield(-name=>'BorderColor',-value=>'none',-size=>15),
    "</td>\n";
  print "</tr>\n";
  print '</table><p>';
  print "</dl>\n";
  print $q->endform, "\n";
  print <<XXX;
XXX
  ;
  Trailer(undef);
}

#
# Display a warning.
#
sub Warning
{
  my($message,$qualifier) = @_;

  Header($message);
  print <<XXX;
<p> <br>
<dl>
<dt><font face="Arial,Helvetica" size=+3>$message:</font>
<p>
<dd><font face="Arial,Helvetica" size=+2>$qualifier</font>
</dl>
<br> <p>
<p>
<br>
<br>
Press a tab above to continue.
XXX
  ;
  Trailer(1);
  die $message;
}

#
# Initialize the CGI context.
#
$CGI::POST_MAX=1024*$MaxFilesize;
$timer=time;
$q=new CGI;
$q->autoEscape(undef);
$q->private_tempfiles(1);
$q->delete('CacheID');
$q->param(-name=>'CacheID',-value=>rand($timer));
if ($q->server_name() =~ /www/)
  {
    print $q->redirect('http://studio.imagemagick.org/MagickStudio/scripts/MagickStudio.cgi');
    exit;
  }
Image::Magick->new('disk-limit'=>$DiskLimit);
#
# Choose function as determined by the query and environment.
#
$header=undef;
$action=$q->param('Action');
$q->delete('Action');
Input() if defined($q->param('File'));
Input() if defined($q->param('URL'));
InputForm() unless defined($action);
Error('You must specify a filename or URL') unless $q->param('Path');
%Functions=
(
  'annotate'=>\&ChooseTool,
  'composite'=>\&ChooseTool,
  'decorate'=>\&ChooseTool,
  'draw'=>\&ChooseTool,
  'effect'=>\&ChooseTool,
  'enhance'=>\&ChooseTool,
  'generate'=>\&ChooseTool,
  'identify'=>\&ChooseTool,
  'mogrify'=>\&Mogrify,
  'Mogrify'=>\&Mogrify,
  'output'=>\&ChooseTool,
  'quantize'=>\&ChooseTool,
  'resize'=>\&ChooseTool,
  'send'=>\&ChooseTool,
  'transform'=>\&ChooseTool,
  'upload'=>\&Upload,
  'view'=>\&ChooseTool,
);
$function=$Functions{$action};
&$function if defined($function);
Error('Request failed due to malformed query');
